from tesserax.animation import Animation, Parallel, KeyframeAnimation
from .core import Body
from .forces import Field
from .constraints import Constraint
from .collisions import Collision


class World:
    def __init__(self):
        self.bodies: list[Body] = []
        self.fields: list[Field] = []
        self.constraints: list[Constraint] = []

    def add(self, shape, **kwargs) -> Body:
        b = Body(shape, **kwargs)
        self.bodies.append(b)
        return b

    def constraint(self, c: Constraint):
        self.constraints.append(c)
        return c

    def simulate(self, duration: float, dt: float = 0.01) -> Animation:
        steps = int(duration / dt)

        # Tracks now include 'rotation'
        tracks = {b: {"tx": {}, "ty": {}, "rotation": {}} for b in self.bodies}

        time = 0.0
        for _ in range(steps):
            t_norm = time / duration if duration > 0 else 0

            # 1. Record
            for b in self.bodies:
                tracks[b]["tx"][t_norm] = b.pos.x
                tracks[b]["ty"][t_norm] = b.pos.y
                tracks[b]["rotation"][t_norm] = b.rotation  # Radians

            # 2. Physics Step
            self._step(dt)
            time += dt

        # 3. Bake
        anims = []
        for b, props in tracks.items():
            anims.append(
                KeyframeAnimation(
                    b.shape,
                    tx=props["tx"],
                    ty=props["ty"],
                    rotation=props["rotation"],  # This maps to transform.rotation
                )
            )

        return Parallel(*anims)

    def _step(self, dt: float):
        # 1. Apply Fields (Gravity, Drag)
        for f in self.fields:
            f.apply(self.bodies)

        # 2. Solve Constraints (Springs apply forces)
        for c in self.constraints:
            c.solve()

        # 3. Integrate (Move objects)
        for b in self.bodies:
            b.integrate(dt)

        # 4. Resolve Collisions (Impulses)
        # Naive O(N^2) - Fine for < 100 objects
        count = len(self.bodies)
        for i in range(count):
            for j in range(i + 1, count):
                a = self.bodies[i]
                b = self.bodies[j]

                # Optimization: Don't check static vs static
                if a.static and b.static:
                    continue

                if col := Collision.solve(a, b):
                    col.resolve()



import math
import heapq
from typing import Iterator
from tesserax.core import Point, Bounds, Shape
from tesserax.base import Group


class Grid:
    def __init__(self, group: Group, size: float = 20.0, limit: int = 10000):
        self.group = group
        self.size = size
        self.limit = limit  # Prevent infinite loops
        self.occupied: set[tuple[int, int]] = set()
        self.bounds_idx: tuple[int, int, int, int] = (0, 0, 0, 0)

        self._rasterize()

    def _to_grid(self, x: float, y: float) -> tuple[int, int]:
        return (math.floor(x / self.size + 0.5), math.floor(y / self.size + 0.5))

    def _to_world(self, gx: int, gy: int) -> Point:
        return Point(gx * self.size, gy * self.size)

    def _rasterize(self):
        self.occupied.clear()

        # Track min/max to define a search bounding box
        min_gx, min_gy = float("inf"), float("inf")
        max_gx, max_gy = float("-inf"), float("-inf")

        for shape in self.group.shapes:
            b = shape.bounds()
            gx1, gy1 = self._to_grid(b.x, b.y)
            gx2, gy2 = self._to_grid(b.x + b.width, b.y + b.height)

            # Update global grid bounds
            min_gx, min_gy = min(min_gx, gx1), min(min_gy, gy1)
            max_gx, max_gy = max(max_gx, gx2), max(max_gy, gy2)

            for gx in range(gx1, gx2 + 1):
                for gy in range(gy1, gy2 + 1):
                    self.occupied.add((gx, gy))

        # Save bounds with generous padding (e.g., 10 cells)
        pad = 10
        self.bounds_idx = (
            int(min_gx - pad),
            int(min_gy - pad),
            int(max_gx + pad),
            int(max_gy + pad),
        )

    def _snap_to_free(
        self, gx: int, gy: int, target_gx: int, target_gy: int
    ) -> tuple[int, int]:
        """
        Finds the nearest free cell to (gx, gy).
        Tie-breaker: Pick the cell closest to (target_gx, target_gy).
        """
        if (gx, gy) not in self.occupied:
            return (gx, gy)

        # Search in expanding rings to ensure we find the strictly nearest cells first
        r = 1
        max_r = 20  # Search radius limit

        while r < max_r:
            candidates = []

            # Iterate only the perimeter of the box at radius r
            # Top and Bottom rows
            for dx in range(-r, r + 1):
                candidates.append((gx + dx, gy - r))
                candidates.append((gx + dx, gy + r))

            # Left and Right columns (excluding corners already added)
            for dy in range(-r + 1, r):
                candidates.append((gx - r, gy + dy))
                candidates.append((gx + r, gy + dy))

            # Filter for valid (free) candidates
            valid_candidates = [c for c in candidates if c not in self.occupied]

            if valid_candidates:
                # HEURISTIC: Choose the candidate with minimum Euclidean distance to the target
                # This biases the snap to move "towards" the destination.
                return min(
                    valid_candidates,
                    key=lambda c: (c[0] - target_gx) ** 2 + (c[1] - target_gy) ** 2,
                )

            r += 1

        return (gx, gy)  # Fail-safe

    def _neighbors(self, gx: int, gy: int) -> Iterator[tuple[int, int]]:
        min_x, min_y, max_x, max_y = self.bounds_idx

        for dx, dy in [(0, 1), (0, -1), (1, 0), (-1, 0)]:
            nx, ny = gx + dx, gy + dy

            # 1. Check Scene Bounds (Stops infinite expansion)
            if not (min_x <= nx <= max_x and min_y <= ny <= max_y):
                continue

            # 2. Check Collision
            if (nx, ny) not in self.occupied:
                yield (nx, ny)

    def _resolve_elbow(
        self, p_user: Point, p_grid: Point, p_next: Point
    ) -> Point | None:
        """
        Calculates an intermediate point (elbow) to ensure orthogonal connection.
        Strategy: Align the elbow segment with the dominant direction of the path.
        """
        # If already aligned, no elbow needed
        if abs(p_user.x - p_grid.x) < 1e-6 or abs(p_user.y - p_grid.y) < 1e-6:
            return p_grid

        # Determine flow direction of the grid path
        dx = p_next.x - p_grid.x
        dy = p_next.y - p_grid.y

        # If path is moving Horizontally, we want to enter/exit horizontally
        # to avoid Z-shapes.
        if abs(dx) > abs(dy):
            # Elbow should have same Y as p_grid, same X as p_user
            return Point(p_user.x, p_grid.y)

        # If path is moving Vertically (or stationary), enter/exit vertically
        else:
            # Elbow should have same X as p_grid, same Y as p_user
            return Point(p_grid.x, p_user.y)

    def trace(self, start: Point, end: Point, *, turn_penalty=1.0) -> list[Point]:
        """A* Pathfinding with safety limits."""
        raw_start = self._to_grid(start.x, start.y)
        raw_end = self._to_grid(end.x, end.y)

        # Fix: Ensure start/end are actually walkable
        start_node = self._snap_to_free(*raw_start, *raw_end)
        end_node = self._snap_to_free(*raw_end, *raw_start)

        # State: (x, y, dx, dy)
        # dx, dy represent the direction we ARRIVED from.
        # For start node, we infer a "virtual" arrival direction based on the User->Grid vector
        # This helps the path align immediately with the user's anchor.
        start_dx = 0
        start_dy = 0

        if start_node != raw_start:
            # If we snapped, the direction is from Raw to Snapped
            start_dx = start_node[0] - raw_start[0]
            start_dy = start_node[1] - raw_start[1]
            # Normalize to -1, 0, 1
            if start_dx != 0:
                start_dx //= abs(start_dx)
            if start_dy != 0:
                start_dy //= abs(start_dy)

        initial_state = (start_node[0], start_node[1], start_dx, start_dy)

        open_set = []
        heapq.heappush(open_set, (0, initial_state))

        came_from = {}
        g_score = {initial_state: 0.0}

        final_state = None
        iterations = 0

        while open_set:
            iterations += 1
            if iterations > self.limit:
                return [start, end]

            _, current = heapq.heappop(open_set)
            cx, cy, cdx, cdy = current

            if (cx, cy) == end_node:
                final_state = current
                break

            for nx, ny in self._neighbors(cx, cy):
                # Calculate new direction
                ndx = nx - cx
                ndy = ny - cy

                # Calculate Cost
                # Base movement cost = 1
                move_cost = 1

                # Turn Penalty
                # If we had a previous direction (cdx!=0 or cdy!=0) and it changed...
                if (cdx != 0 or cdy != 0) and (ndx != cdx or ndy != cdy):
                    move_cost += turn_penalty

                new_g = g_score[current] + move_cost
                next_state = (nx, ny, ndx, ndy)

                if next_state not in g_score or new_g < g_score[next_state]:
                    g_score[next_state] = new_g

                    # Heuristic (Manhattan)
                    h = abs(end_node[0] - nx) + abs(end_node[1] - ny)

                    heapq.heappush(open_set, (new_g + h, next_state))
                    came_from[next_state] = current

        if not final_state:
            return [start, end]

        # Reconstruct path
        path = []
        curr = final_state
        while curr in came_from:
            path.append((curr[0], curr[1]))
            curr = came_from[curr]
        path.append(start_node)
        path.reverse()

        # Simplify Path (Collinear Check)
        if len(path) < 3:
            return [start, end]

        simplified = [self._to_world(*path[0])]  # Use exact start point
        last_dir = (path[1][0] - path[0][0], path[1][1] - path[0][1])

        for i in range(2, len(path)):
            curr_dir = (path[i][0] - path[i - 1][0], path[i][1] - path[i - 1][1])
            if curr_dir != last_dir:
                simplified.append(self._to_world(*path[i - 1]))
                last_dir = curr_dir

        simplified.append(self._to_world(*path[-1]))  # Use exact end point

        if len(simplified) == 1:
            simplified = simplified * 2

        return (
            [start, self._resolve_elbow(start, simplified[0], simplified[1])]
            + simplified[1:-1]
            + [self._resolve_elbow(end, simplified[-1], simplified[-2]), end]
        )



"""
This module defines the `Color` class with methods to create and manipulate
colors in different spaces. It also defines all common CSS colors, and utilities
to combine them.
"""

from __future__ import annotations
import colorsys
from dataclasses import dataclass
import sys
from binascii import hexlify
from typing import Any, Generator, List, Tuple, overload


def scale(unit, min_value, max_value, steps):
    current = min_value
    delta = (max_value - min_value) / (steps - 1)

    for _ in range(steps):
        yield unit(current)
        current += delta


# ## The `Color` class


@dataclass
class Color:
    r: int = 0
    g: int = 0
    b: int = 0
    a: float = 0.0

    def __str__(self):
        return f"rgba({self.r},{self.g},{self.b},{self.a})"

    def __repr__(self):
        return f"Color({self.r},{self.g},{self.b}, alpha={self.a})"

    def __eq__(self, __o: object) -> bool:
        return isinstance(__o, Color) and repr(self) == repr(__o)

    def __hash__(self) -> int:
        return hash(repr(self))

    @property
    def name(self) -> str:
        return Colors.get_name(self)

    @property
    def lightness(self):
        return hls(self)[1]

    @property
    def hue(self):
        return hls(self)[0]

    @property
    def value(self):
        return hsv(self)[2]

    @property
    def saturation(self):
        return hls(self)[2]

    def saturated(self, saturation: float) -> Color:
        h, _, v = hsv(self)
        return hsv(h, saturation, v, alpha=self.a)

    def lit(self, lightness: float) -> Color:
        h, _, s = hls(self)
        return hls(h, lightness, s, alpha=self.a)

    def shifted(self, hue: float) -> Color:
        _, l, s = hls(self)
        return hls(hue, l, s, alpha=self.a)

    def transparent(self, alpha: float) -> Color:
        return Color(self.r, self.g, self.b, alpha)

    def lighter(self, alpha: float) -> Color:
        return self.lerp(Colors.White, alpha, space="rgb")

    def darker(self, alpha: float) -> Color:
        return self.lerp(Colors.Black, alpha, space="rgb")

    def brighter(self, alpha: float) -> Color:
        return self.lerp(self.saturated(1.0), alpha, space="hsv")

    def dimmer(self, alpha: float) -> Color:
        return self.lerp(self.saturated(0.0), alpha, space="hsv")

    def redshift(self, percent: float) -> Color:
        _, l, s = hls(self)
        target = hls(0, l, s)
        return self.lerp(target, percent, space="rgb")

    def blueshift(self, percent: float) -> Color:
        _, l, s = hls(self)
        target = hls(1, l, s)
        return self.lerp(target, percent, space="rgb")

    def lerp(self, other: Color, percent: float, *, space="hls") -> Color:
        space = dict(rgb=rgb, hls=hls, hsv=hsv)[space]

        start_values = space(self)
        end_values = space(other)
        result = [s + (e - s) * percent for s, e in zip(start_values, end_values)]

        return space(*result)

    def shade(self, value: float) -> Color:
        middle = self

        if value < 0.5:
            # 0..0.5 -> 1 ... 0
            return middle.darker(1 - 2 * value)

        if value > 0.5:
            # 0.5..1 -> 0 ... 1
            return middle.lighter(2 * value - 1)

        return middle

    @staticmethod
    def palette(start: Color, end: Color, steps: int, space="hls") -> List[Color]:
        percents = scale(float, 0, 1, steps)
        return [start.lerp(end, p, space=space) for p in percents]

    def __add__(self, other: Color) -> Color:
        return rgb(
            self.r / 255 + other.r / 255,
            self.g / 255 + other.g / 255,
            self.b / 255 + other.b / 255,
            alpha=self.a
            + other.a,  # Additive alpha? Usually mix, but for vectors + is correct
        )

    def __sub__(self, other: Color) -> Color:
        return rgb(
            self.r / 255 - other.r / 255,
            self.g / 255 - other.g / 255,
            self.b / 255 - other.b / 255,
            alpha=self.a - other.a,
        )

    def __mul__(self, other: int | float) -> Color:
        return rgb(
            self.r / 255 * other,
            self.g / 255 * other,
            self.b / 255 * other,
            alpha=self.a * other,
        )


# ## Color spaces


@overload
def rgb(red: float, green: float, blue: float, /, *, alpha: float = 1.0) -> Color: ...


@overload
def rgb(color: Color, /) -> Tuple[float, float, float]: ...


def rgb(*args, **kwargs):
    if isinstance(args[0], Color):
        color = args[0]
        return color.r / 255, color.g / 255, color.b / 255
    else:
        r, g, b = args
        alpha = kwargs.pop("alpha", 1.0)

        return Color(int(r * 255), int(g * 255), int(b * 255), alpha)


@overload
def hsv(
    hue: float, saturation: float, value: float, /, *, alpha: float = 1.0
) -> Color: ...


@overload
def hsv(color: Color, /) -> Tuple[float, float, float]: ...


def hsv(*args, **kwargs):
    if isinstance(args[0], Color):
        color = args[0]
        return colorsys.rgb_to_hsv(*rgb(color))
    else:
        h, s, v = args
        r, g, b = colorsys.hsv_to_rgb(h, s, v)
        alpha = kwargs.pop("alpha", 1.0)

        return rgb(r, g, b, alpha=alpha)


@overload
def hls(
    hue: float, lightness: float, saturation: float, /, *, alpha: float = 1.0
) -> Color: ...


@overload
def hls(color: Color, /) -> Tuple[float, float, float]: ...


def hls(*args, **kwargs):
    if isinstance(args[0], Color):
        color = args[0]
        return colorsys.rgb_to_hls(*rgb(color))
    else:
        h, l, s = args
        r, g, b = colorsys.hls_to_rgb(h, l, s)
        alpha = kwargs.pop("alpha", 1.0)

        return rgb(r, g, b, alpha=alpha)


@overload
def hex(code: str, /, *, alpha: float = 1.0) -> Color: ...


@overload
def hex(color: Color, /) -> str: ...


def hex(*args, **kwargs):
    if isinstance(args[0], str):
        code = args[0]
        alpha = kwargs.pop("alpha", 1.0)

        if code.startswith("#"):
            code = code[1:]

        if len(code) == 3:
            r = code[0] * 2
            g = code[1] * 2
            b = code[2] * 2
        else:
            r = code[0:2]
            g = code[2:4]
            b = code[4:6]

        return Color(int(r, 16), int(g, 16), int(b, 16), alpha)
    else:
        color: Color = args[0]

        r = hexlify(color.r.to_bytes(1, sys.byteorder)).decode("ascii")
        g = hexlify(color.g.to_bytes(1, sys.byteorder)).decode("ascii")
        b = hexlify(color.b.to_bytes(1, sys.byteorder)).decode("ascii")
        return f"#{r}{g}{b}"


# ## Basic color shorthands


def red(lightness: float = 1.0) -> Color:
    return Colors.Red.shade(lightness)


def green(lightness: float = 1.0) -> Color:
    return Colors.Green.shade(lightness)


def blue(lightness: float = 1.0) -> Color:
    return Colors.Blue.shade(lightness)


def gray(lightness: float = 0.5) -> Color:
    return Colors.Gray.shade(lightness)


# ## The `Colors` class

# These are the web colors defined in CSS.
# They are defined as class members of the `Colors` class,
# so they can be accessed by dot notation.

# Similarly, the `Colors` class provides an item getter
# to access colors by name.


class Colors:
    """Contains class-member definitions for all standard CSS colors.

    Colors can be accessed by dot notation (`Colors.Lime`) or
    by name (`Colors['Fuchsia']`).
    """

    __color_names = {}
    __colors = {}

    @classmethod
    def get(cls, name: str) -> Color:
        try:
            return cls.__colors[name]
        except KeyError:
            raise ValueError(f"No color named {repr(name)}.")

    @classmethod
    def get_name(cls, color: Color) -> str:
        try:
            return Colors.__color_names[color]
        except KeyError:
            raise ValueError(f"Color {repr(color)} not registered.")

    @classmethod
    def define_color(cls, color: Color, name: str):
        cls.__color_names[color] = name
        cls.__colors[name] = color
        cls.__colors[name.lower()] = color

    # Transparent
    Transparent = Color()

    # Basic web colors
    White = hex("#ffffff")
    Silver = hex("#c0c0c0")
    Gray = hex("#808080")
    Black = hex("#000000")
    Red = hex("#ff0000")
    Maroon = hex("#800000")
    Yellow = hex("#ffff00")
    Olive = hex("#808000")
    Lime = hex("#00ff00")
    Green = hex("#008000")
    Aqua = hex("#00ffff")
    Teal = hex("#008080")
    Blue = hex("#0000ff")
    Navy = hex("#000080")
    Fuchsia = hex("#ff00ff")
    Purple = hex("#800080")

    # Pink shades
    MediumVioletRed = hex("#c71585")
    DeepPink = hex("#ff1493")
    PaleVioletRed = hex("#db7093")
    HotPink = hex("#ff69b4")
    LightPink = hex("#ffb6c1")
    Pink = hex("#ffc0cb")

    # Red shades
    DarkRed = hex("#8b0000")
    Red = hex("#ff0000")
    Firebrick = hex("#b22222")
    Crimson = hex("#dc143c")
    IndianRed = hex("#cd5c5c")
    LightCoral = hex("#f08080")
    Salmon = hex("#fa8072")
    DarkSalmon = hex("#e9967a")
    LightSalmon = hex("#ffa07a")

    # Orange shades
    OrangeRed = hex("#ff4500")
    Tomato = hex("#ff6347")
    DarkOrange = hex("#ff8c00")
    Coral = hex("#ff7f50")
    Orange = hex("#ffa500")

    # Yellow shades
    DarkKhaki = hex("#bdb76b")
    Gold = hex("#ffd700")
    Khaki = hex("#f0e68c")
    PeachPuff = hex("#ffdab9")
    Yellow = hex("#ffff00")
    PaleGoldenrod = hex("#eee8aa")
    Moccasin = hex("#ffe4b5")
    PapayaWhip = hex("#ffefd5")
    LightGoldenrodYellow = hex("#fafad2")
    LemonChiffon = hex("#fffacd")
    LightYellow = hex("#ffffe0")

    # Brown shades
    Maroon = hex("#800000")
    Brown = hex("#a52a2a")
    SaddleBrown = hex("#8b4513")
    Sienna = hex("#a0522d")
    Chocolate = hex("#d2691e")
    DarkGoldenrod = hex("#b8860b")
    Peru = hex("#cd853f")
    RosyBrown = hex("#bc8f8f")
    Goldenrod = hex("#daa520")
    SandyBrown = hex("#f4a460")
    Tan = hex("#d2b48c")
    Burlywood = hex("#deb887")
    Wheat = hex("#f5deb3")
    NavajoWhite = hex("#ffdead")
    Bisque = hex("#ffe4c4")
    BlanchedAlmond = hex("#ffebcd")
    Cornsilk = hex("#fff8dc")

    # Green shades
    DarkGreen = hex("#006400")
    Green = hex("#008000")
    DarkOliveGreen = hex("#556b2f")
    ForestGreen = hex("#228b22")
    SeaGreen = hex("#2e8b57")
    Olive = hex("#808000")
    OliveDrab = hex("#6b8e23")
    MediumSeaGreen = hex("#3cb371")
    LimeGreen = hex("#32cd32")
    Lime = hex("#00ff00")
    SpringGreen = hex("#00ff7f")
    MediumSpringGreen = hex("#00fa9a")
    DarkSeaGreen = hex("#8fbc8f")
    MediumAquamarine = hex("#66cdaa")
    YellowGreen = hex("#9acd32")
    LawnGreen = hex("#7cfc00")
    Chartreuse = hex("#7fff00")
    LightGreen = hex("#90ee90")
    GreenYellow = hex("#adff2f")
    PaleGreen = hex("#98fb98")

    # Cyan shades
    Teal = hex("#008080")
    DarkCyan = hex("#008b8b")
    LightSeaGreen = hex("#20b2aa")
    CadetBlue = hex("#5f9ea0")
    DarkTurquoise = hex("#00ced1")
    MediumTurquoise = hex("#48d1cc")
    Turquoise = hex("#40e0d0")
    Aqua = hex("#00ffff")
    Cyan = hex("#00ffff")
    Aquamarine = hex("#7fffd4")
    PaleTurquoise = hex("#afeeee")
    LightCyan = hex("#e0ffff")

    # Blue shades
    MidnightBlue = hex("#191970")
    Navy = hex("#000080")
    DarkBlue = hex("#00008b")
    MediumBlue = hex("#0000cd")
    Blue = hex("#0000ff")
    RoyalBlue = hex("#4169e1")
    SteelBlue = hex("#4682b4")
    DodgerBlue = hex("#1e90ff")
    DeepSkyBlue = hex("#00bfff")
    CornflowerBlue = hex("#6495ed")
    SkyBlue = hex("#87ceeb")
    LightSkyBlue = hex("#87cefa")
    LightSteelBlue = hex("#b0c4de")
    LightBlue = hex("#add8e6")
    PowderBlue = hex("#b0e0e6")

    # Purple shades
    Indigo = hex("#4b0082")
    Purple = hex("#800080")
    DarkMagenta = hex("#8b008b")
    DarkViolet = hex("#9400d3")
    DarkSlateBlue = hex("#483d8b")
    BlueViolet = hex("#8a2be2")
    DarkOrchid = hex("#9932cc")
    Fuchsia = hex("#ff00ff")
    Magenta = hex("#ff00ff")
    SlateBlue = hex("#6a5acd")
    MediumSlateBlue = hex("#7b68ee")
    MediumOrchid = hex("#ba55d3")
    MediumPurple = hex("#9370db")
    Orchid = hex("#da70d6")
    Violet = hex("#ee82ee")
    Plum = hex("#dda0dd")
    Thistle = hex("#d8bfd8")
    Lavender = hex("#e6e6fa")

    # White shades
    MistyRose = hex("#ffe4e1")
    AntiqueWhite = hex("#faebd7")
    Linen = hex("#faf0e6")
    Beige = hex("#f5f5dc")
    WhiteSmoke = hex("#f5f5f5")
    LavenderBlush = hex("#fff0f5")
    OldLace = hex("#fdf5e6")
    AliceBlue = hex("#f0f8ff")
    Seashell = hex("#fff5ee")
    GhostWhite = hex("#f8f8ff")
    Honeydew = hex("#f0fff0")
    FloralWhite = hex("#fffaf0")
    Azure = hex("#f0ffff")
    MintCream = hex("#f5fffa")
    Snow = hex("#fffafa")
    Ivory = hex("#fffff0")
    White = hex("#ffffff")

    # Black shades
    Black = hex("#000000")
    DarkSlateGray = hex("#2f4f4f")
    DimGray = hex("#696969")
    SlateGray = hex("#708090")
    Gray = hex("#808080")
    LightSlateGray = hex("#778899")
    DarkGray = hex("#a9a9a9")
    Silver = hex("#c0c0c0")
    LightGray = hex("#d3d3d3")
    Gainsboro = hex("#dcdcdc")

    # Extra
    RebeccaPurple = hex("#663399")

    # Palettes

    @staticmethod
    def basic_palette():
        yield Colors.White
        yield Colors.Silver
        yield Colors.Gray
        yield Colors.Black
        yield Colors.Red
        yield Colors.Maroon
        yield Colors.Yellow
        yield Colors.Olive
        yield Colors.Lime
        yield Colors.Green
        yield Colors.Aqua
        yield Colors.Teal
        yield Colors.Blue
        yield Colors.Navy
        yield Colors.Fuchsia
        yield Colors.Purple

    @staticmethod
    def pink_palette():
        yield Colors.MediumVioletRed
        yield Colors.DeepPink
        yield Colors.PaleVioletRed
        yield Colors.HotPink
        yield Colors.LightPink
        yield Colors.Pink

    @staticmethod
    def red_palette():
        yield Colors.DarkRed
        yield Colors.Red
        yield Colors.Firebrick
        yield Colors.Crimson
        yield Colors.IndianRed
        yield Colors.LightCoral
        yield Colors.Salmon
        yield Colors.DarkSalmon
        yield Colors.LightSalmon

    @staticmethod
    def orange_palette():
        yield Colors.OrangeRed
        yield Colors.Tomato
        yield Colors.DarkOrange
        yield Colors.Coral
        yield Colors.Orange

    @staticmethod
    def yellow_palette():
        yield Colors.DarkKhaki
        yield Colors.Gold
        yield Colors.Khaki
        yield Colors.PeachPuff
        yield Colors.Yellow
        yield Colors.PaleGoldenrod
        yield Colors.Moccasin
        yield Colors.PapayaWhip
        yield Colors.LightGoldenrodYellow
        yield Colors.LemonChiffon
        yield Colors.LightYellow

    @staticmethod
    def brown_palette():
        yield Colors.Maroon
        yield Colors.Brown
        yield Colors.SaddleBrown
        yield Colors.Sienna
        yield Colors.Chocolate
        yield Colors.DarkGoldenrod
        yield Colors.Peru
        yield Colors.RosyBrown
        yield Colors.Goldenrod
        yield Colors.SandyBrown
        yield Colors.Tan
        yield Colors.Burlywood
        yield Colors.Wheat
        yield Colors.NavajoWhite
        yield Colors.Bisque
        yield Colors.BlanchedAlmond
        yield Colors.Cornsilk

    @staticmethod
    def green_palette():
        yield Colors.DarkGreen
        yield Colors.Green
        yield Colors.DarkOliveGreen
        yield Colors.ForestGreen
        yield Colors.SeaGreen
        yield Colors.Olive
        yield Colors.OliveDrab
        yield Colors.MediumSeaGreen
        yield Colors.LimeGreen
        yield Colors.Lime
        yield Colors.SpringGreen
        yield Colors.MediumSpringGreen
        yield Colors.DarkSeaGreen
        yield Colors.MediumAquamarine
        yield Colors.YellowGreen
        yield Colors.LawnGreen
        yield Colors.Chartreuse
        yield Colors.LightGreen
        yield Colors.GreenYellow
        yield Colors.PaleGreen

    @staticmethod
    def cyan_palette():
        yield Colors.Teal
        yield Colors.DarkCyan
        yield Colors.LightSeaGreen
        yield Colors.CadetBlue
        yield Colors.DarkTurquoise
        yield Colors.MediumTurquoise
        yield Colors.Turquoise
        yield Colors.Aqua
        yield Colors.Cyan
        yield Colors.Aquamarine
        yield Colors.PaleTurquoise
        yield Colors.LightCyan

    @staticmethod
    def blue_palette():
        yield Colors.MidnightBlue
        yield Colors.Navy
        yield Colors.DarkBlue
        yield Colors.MediumBlue
        yield Colors.Blue
        yield Colors.RoyalBlue
        yield Colors.SteelBlue
        yield Colors.DodgerBlue
        yield Colors.DeepSkyBlue
        yield Colors.CornflowerBlue
        yield Colors.SkyBlue
        yield Colors.LightSkyBlue
        yield Colors.LightSteelBlue
        yield Colors.LightBlue
        yield Colors.PowderBlue

    @staticmethod
    def purple_palette():
        yield Colors.Indigo
        yield Colors.Purple
        yield Colors.DarkMagenta
        yield Colors.DarkViolet
        yield Colors.DarkSlateBlue
        yield Colors.BlueViolet
        yield Colors.DarkOrchid
        yield Colors.Fuchsia
        yield Colors.Magenta
        yield Colors.SlateBlue
        yield Colors.MediumSlateBlue
        yield Colors.MediumOrchid
        yield Colors.MediumPurple
        yield Colors.Orchid
        yield Colors.Violet
        yield Colors.Plum
        yield Colors.Thistle
        yield Colors.Lavender

    @staticmethod
    def white_palette():
        yield Colors.MistyRose
        yield Colors.AntiqueWhite
        yield Colors.Linen
        yield Colors.Beige
        yield Colors.WhiteSmoke
        yield Colors.LavenderBlush
        yield Colors.OldLace
        yield Colors.AliceBlue
        yield Colors.Seashell
        yield Colors.GhostWhite
        yield Colors.Honeydew
        yield Colors.FloralWhite
        yield Colors.Azure
        yield Colors.MintCream
        yield Colors.Snow
        yield Colors.Ivory
        yield Colors.White

    @staticmethod
    def black_palette():
        yield Colors.Black
        yield Colors.DarkSlateGray
        yield Colors.DimGray
        yield Colors.SlateGray
        yield Colors.Gray
        yield Colors.LightSlateGray
        yield Colors.DarkGray
        yield Colors.Silver
        yield Colors.LightGray
        yield Colors.Gainsboro

    @staticmethod
    def extra_palette():
        yield Colors.RebeccaPurple

    @staticmethod
    def all():
        yield from Colors.pink_palette()
        yield from Colors.red_palette()
        yield from Colors.orange_palette()
        yield from Colors.yellow_palette()
        yield from Colors.brown_palette()
        yield from Colors.green_palette()
        yield from Colors.cyan_palette()
        yield from Colors.blue_palette()
        yield from Colors.purple_palette()
        yield from Colors.white_palette()
        yield from Colors.black_palette()
        yield from Colors.extra_palette()

    @classmethod
    def palette(cls, palette: str) -> Generator[Color, Any, Any]:
        try:
            return getattr(cls, f"{palette}_palette")()
        except:
            raise ValueError(f"Palette {palette} doesn't exist.")

    @staticmethod
    def palettes():
        return [
            "pink",
            "red",
            "orange",
            "yellow",
            "brown",
            "green",
            "cyan",
            "blue",
            "purple",
            "white",
            "black",
            "extra",
        ]


for name in dir(Colors):
    obj = getattr(Colors, name)

    if isinstance(obj, Color):
        Colors.define_color(obj, name)



from __future__ import annotations
import copy
from dataclasses import dataclass
from abc import ABC, abstractmethod
import math
from typing import Literal, Protocol, Self, TYPE_CHECKING
from .color import Color, Colors

if TYPE_CHECKING:
    from .base import Group
    from .animation import Animator

type Anchor = Literal[
    "top",
    "bottom",
    "left",
    "right",
    "center",
    "topleft",
    "topright",
    "bottomleft",
    "bottomright",
]


def deg(degrees: float) -> float:
    """Use this method to convert to radians."""
    return math.radians(degrees)


@dataclass(frozen=True)
class Point:
    x: float
    y: float

    @classmethod
    def zero(cls) -> Point:
        return cls(0, 0)

    @classmethod
    def up(cls) -> Point:
        return cls(0, -1)  # SVG Y is down

    @classmethod
    def down(cls) -> Point:
        return cls(0, 1)

    @classmethod
    def left(cls) -> Point:
        return cls(-1, 0)

    @classmethod
    def right(cls) -> Point:
        return cls(1, 0)

    @staticmethod
    def distance_to_segment(p: Point, start: Point, end: Point) -> float:
        # Area of triangle = 0.5 * base * height
        # Height = (2 * Area) / base
        # Area can be found via cross product
        dx = end.x - start.x
        dy = end.y - start.y

        if dx == 0 and dy == 0:
            return (p - start).magnitude()

        num = abs(dy * p.x - dx * p.y + end.x * start.y - end.y * start.x)
        den = math.sqrt(dy**2 + dx**2)
        return num / den

    def apply(self, tx=0.0, ty=0.0, r=0.0, s=1.0) -> Point:
        rad = math.radians(r)
        nx, ny = self.x * s, self.y * s
        rx = nx * math.cos(rad) - ny * math.sin(rad)
        ry = nx * math.sin(rad) + ny * math.cos(rad)
        return Point(rx + tx, ry + ty)

    def __add__(self, other: Point) -> Point:
        return Point(self.x + other.x, self.y + other.y)

    def __sub__(self, other: Point) -> Point:
        return Point(self.x - other.x, self.y - other.y)

    def magnitude(self) -> float:
        return math.sqrt(self.x**2 + self.y**2)

    def normalize(self) -> Point:
        m = self.magnitude()

        if m == 0:
            return Point(0, 0)

        return Point(self.x / m, self.y / m)

    def __mul__(self, scalar: float) -> Point:
        return Point(self.x * scalar, self.y * scalar)

    def __truediv__(self, scalar: float) -> Point:
        return Point(self.x / scalar, self.y / scalar)

    def dx(self, dx: float) -> Point:
        return self + Point(dx, 0)

    def dy(self, dy: float) -> Point:
        return self + Point(0, dy)

    def d(self, dx: float, dy: float) -> Point:
        return self + Point(dx, dy)

    def lerp(self, other: Point, t: float) -> Point:
        """Linear interpolation between this point and another."""
        return self + (other - self) * t


@dataclass
class Transform:
    """
    Represents an affine transformation decomposed into:
    Translation (tx, ty), Rotation (radians), and Scale (sx, sy).
    """

    tx: float = 0.0
    ty: float = 0.0
    rotation: float = 0.0
    sx: float = 1.0
    sy: float = 1.0

    # --- Factory & Lifecycle ---

    @classmethod
    def identity(cls) -> Self:
        return cls()

    def copy(self) -> Transform:
        return Transform(self.tx, self.ty, self.rotation, self.sx, self.sy)

    def reset(self) -> Self:
        self.tx = 0.0
        self.ty = 0.0
        self.rotation = 0.0
        self.sx = 1.0
        self.sy = 1.0
        return self

    # --- Animation Logic ---

    def lerp(self, target: "Transform", t: float) -> "Transform":
        """
        Linearly interpolates between this transform and a target.
        Returns a NEW Transform instance.
        """
        # Clamp t to [0, 1] to prevent overshooting
        t = max(0.0, min(1.0, t))

        return Transform(
            tx=self.tx + (target.tx - self.tx) * t,
            ty=self.ty + (target.ty - self.ty) * t,
            # Interpolating angle directly preserves rotational direction
            rotation=self.rotation + (target.rotation - self.rotation) * t,
            sx=self.sx + (target.sx - self.sx) * t,
            sy=self.sy + (target.sy - self.sy) * t,
        )

    # --- Math & Application ---

    def map(self, p: "Point") -> "Point":
        """
        Applies the transformation to a Point (Scale -> Rotate -> Translate).
        """
        # 1. Scale
        x = p.x * self.sx
        y = p.y * self.sy

        # 2. Rotate
        if self.rotation != 0:
            cos_theta = math.cos(self.rotation)
            sin_theta = math.sin(self.rotation)
            x_new = x * cos_theta - y * sin_theta
            y_new = x * sin_theta + y * cos_theta
            x, y = x_new, y_new

        # 3. Translate
        return p.__class__(x + self.tx, y + self.ty)

    # --- Fluent Mutators ---

    def translate(self, dx: float, dy: float) -> Self:
        self.tx += dx
        self.ty += dy
        return self

    def rotate(self, radians: float) -> Self:
        self.rotation += radians
        return self

    def scale(self, sx: float, sy: float | None = None) -> Self:
        if sy is None:
            sy = sx

        self.sx *= sx
        self.sy *= sy
        return self


@dataclass(frozen=True)
class Bounds:
    x: float
    y: float
    width: float
    height: float

    @property
    def left(self) -> Point:
        return Point(self.x, self.y + self.height / 2)

    @property
    def right(self) -> Point:
        return Point(self.x + self.width, self.y + self.height / 2)

    @property
    def top(self) -> Point:
        return Point(self.x + self.width / 2, self.y)

    @property
    def bottom(self) -> Point:
        return Point(self.x + self.width / 2, self.y + self.height)

    @property
    def topleft(self) -> Point:
        return Point(self.x, self.y)

    @property
    def topright(self) -> Point:
        return Point(self.x + self.width, self.y)

    @property
    def bottomleft(self) -> Point:
        return Point(self.x, self.y + self.height)

    @property
    def bottomright(self) -> Point:
        return Point(self.x + self.width, self.y + self.height)

    @property
    def center(self) -> Point:
        return Point(self.x + self.width / 2, self.y + self.height / 2)

    def padded(self, amount: float) -> Bounds:
        return Bounds(
            self.x - amount,
            self.y - amount,
            self.width + 2 * amount,
            self.height + 2 * amount,
        )

    def anchor(self, name: Anchor) -> Point:
        match name:
            case "top":
                return self.top
            case "bottom":
                return self.bottom
            case "left":
                return self.left
            case "right":
                return self.right
            case "center":
                return self.center
            case "topleft":
                return self.topleft
            case "topright":
                return self.topright
            case "bottomleft":
                return self.bottomleft
            case "bottomright":
                return self.bottomright
            case _:
                raise ValueError(f"Unknown anchor: {name}")

    @classmethod
    def union(cls, *bounds: Bounds) -> Bounds:
        if not bounds:
            return Bounds(0, 0, 0, 0)

        x_min = min(b.x for b in bounds)
        y_min = min(b.y for b in bounds)
        x_max = max(b.x + b.width for b in bounds)
        y_max = max(b.y + b.height for b in bounds)

        return Bounds(x_min, y_min, x_max - x_min, y_max - y_min)


class IShape(Protocol):
    transform: Transform

    def render(self) -> str: ...


class Shape(ABC):
    def __init__(
        self,
    ) -> None:
        from .base import Group
        from .animation import Animator

        self.transform = Transform.identity()
        self.parent: Group | None = None
        self.hidden: bool = False

        if (gp := Group.current()) is not None:
            gp.append(self)

        self._animator: Animator | None = None

    def detach(self) -> Self:
        if self.parent:
            self.parent.remove(self)

        return self

    def attach(self) -> Self:
        from .base import Group

        self.detach()

        if Group.stack:
            Group.stack[-1].append(self)

        return self

    def show(self) -> Self:
        self.hidden = False
        return self

    def hide(self) -> Self:
        self.hidden = True
        return self

    @property
    def animate(self):
        from .animation import Animator

        if self._animator is None:
            self._animator = Animator(self)

        return self._animator

    @abstractmethod
    def local(self) -> Bounds:
        pass

    def bounds(self) -> Bounds:
        base = self.local()

        corners = [base.topleft, base.topright, base.bottomleft, base.bottomright]
        transformed = [self.transform.map(p) for p in corners]
        xs = [p.x for p in transformed]
        ys = [p.y for p in transformed]

        return Bounds(min(xs), min(ys), max(xs) - min(xs), max(ys) - min(ys))

    @abstractmethod
    def _render(self) -> str:
        pass

    def render(self) -> str:
        """Generates the SVG expression for this shape, including transformation."""
        if self.hidden:
            return ""

        t = self.transform

        # Optimization: Return raw render if transform is identity
        # (This avoids nested <g> tags for shapes that haven't moved)
        if t.tx == 0 and t.ty == 0 and t.rotation == 0 and t.sx == 1 and t.sy == 1:
            return self._render()

        # Fix: SVG rotate expects degrees, but Transform stores radians
        deg = math.degrees(t.rotation)

        ts = f' transform="translate({t.tx} {t.ty}) rotate({deg}) scale({t.sx} {t.sy})"'
        return f"<g{ts}>\n{self._render()}\n</g>"

    def resolve(self, p: Point) -> Point:
        world_p = self.transform.map(p)
        if self.parent:
            return self.parent.resolve(world_p)
        return world_p

    def anchor(self, name: Anchor) -> Point:
        return self.resolve(self.local().anchor(name))

    def translated(self, dx: float, dy: float) -> Self:
        self.transform.translate(dx, dy)
        return self

    def rotated(self, r: float) -> Self:
        self.transform.rotate(r)
        return self

    def scaled(self, s: float) -> Self:
        self.transform.scale(s)
        return self

    def clone(self) -> Self:
        return copy.deepcopy(self).attach()

    def __add__(self, other: Shape) -> Group:
        # Import internally to avoid circular import with base.py
        from .base import Group

        return Group().add(self, other)

    def align_to(
        self,
        other: Shape,
        anchor: Anchor,
        other_anchor: Anchor | None = None,
    ) -> Self:
        """
        Aligns this shape to another shape.

        Moves this shape so that its 'anchor' point coincides with
        'other_anchor' on the target shape.
        """
        target_anchor = other_anchor or anchor
        return self.move_to(other.anchor(target_anchor), anchor)

    def move_to(self, target: Point, anchor: Anchor = "center") -> Self:
        """
        Moves the shape so that its specified anchor is at the target Point.
        """
        p_self = self.anchor(anchor)
        diff = target - p_self

        self.transform.tx += diff.x
        self.transform.ty += diff.y

        return self


class Component(Shape):
    """
    A Shape that is composed of other Shapes.
    Subclasses must implement build() which returns the geometric representation.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__()
        self._primitive: Shape | None = None
        self._dynamic = False
        self._kwargs = kwargs

    @abstractmethod
    def _build(self) -> Shape:
        """
        Constructs the visual representation of this component.
        Returns a Shape (e.g., Group, Path, Rect).
        """
        pass

    def refresh(self) -> Self:
        """Invalidates the cache, forcing a rebuild on next use."""
        self._primitive = None
        return self

    @property
    def _shape(self) -> Shape:
        if self._primitive is None or self._dynamic:
            self._primitive = self._build().detach()

        return self._primitive

    def local(self) -> Bounds:
        return self._shape.bounds()

    def _render(self) -> str:
        # Render the built primitive.
        # Note: The primitive's own transform is applied inside its .render()
        # The Component's transform is applied by the caller (Shape.render)
        return self._shape.render()



import math
from tesserax.core import Point, Shape
from tesserax.base import Circle, Rect
from tesserax.physics.colliders import CircleCollider
from tesserax.physics.core import Body
from tesserax.physics.collisions import Collision

# --- Helpers ---


def create_body(shape_cls, *args, pos=(0, 0), rotation=0, **kwargs):
    """Factory to create a physics body at a specific location/rotation."""
    shape: Shape = shape_cls(*args, **kwargs)

    shape.move_to(Point(*pos))

    if rotation != 0:
        shape.rotated(math.radians(rotation))

    collider = CircleCollider(*args, **kwargs) if shape_cls is Circle else None
    return Body(shape, collider=collider)


def assert_collides(a: Body, b: Body):
    """Asserts that A and B logicially intersect."""
    collision = Collision.solve(a, b)
    assert (
        collision is not None
    ), f"Expected collision between {a} and {b}, but found None."


def assert_separate(a: Body, b: Body):
    """Asserts that A and B are logically separated."""
    collision = Collision.solve(a, b)
    assert collision is None, f"Expected no collision, but found {collision}."


# --- 1. Circle vs Circle Logic ---


def test_circles_intersecting():
    # Two circles (r=10) placed 10 units apart.
    # They should overlap significantly.
    a = create_body(Circle, 10, pos=(0, 0))
    b = create_body(Circle, 10, pos=(10, 0))
    assert_collides(a, b)


def test_circles_separated():
    # Two circles (r=10) placed 100 units apart.
    a = create_body(Circle, 10, pos=(0, 0))
    b = create_body(Circle, 10, pos=(100, 0))
    assert_separate(a, b)


def test_circles_touching_barely():
    # Radius 10. Centers at 0 and 19. Overlap is 1.0.
    # Should collide.
    a = create_body(Circle, 10, pos=(0, 0))
    b = create_body(Circle, 10, pos=(19, 0))
    assert_collides(a, b)


# --- 2. Box vs Box Logic ---


def test_boxes_intersecting_aabb():
    # Two 20x20 boxes overlapping.
    a = create_body(Rect, 20, 20, pos=(0, 0))
    b = create_body(Rect, 20, 20, pos=(10, 5))
    assert_collides(a, b)


def test_boxes_separated_aabb():
    # Two 20x20 boxes far apart.
    a = create_body(Rect, 20, 20, pos=(0, 0))
    b = create_body(Rect, 20, 20, pos=(50, 50))
    assert_separate(a, b)


def test_boxes_intersecting_rotated():
    # Box A is axis-aligned.
    # Box B is rotated 45 degrees and its corner pokes into A.
    a = create_body(Rect, 100, 100, pos=(0, 0))

    # Place B such that its center is outside, but rotation makes it hit.
    # Center X=60 (outside A's extent of 50).
    # But rotated corner should reach back into A.
    b = create_body(Rect, 40, 40, pos=(60, 0), rotation=45)

    assert_collides(a, b)


def test_boxes_separated_rotated():
    # Same setup, but B is moved further away so even the corner doesn't reach.
    a = create_body(Rect, 100, 100, pos=(0, 0))
    b = create_body(Rect, 40, 40, pos=(100, 0), rotation=45)
    assert_separate(a, b)


# --- 3. Circle vs Box Logic ---


def test_circle_inside_box():
    # Circle completely swallowed by box.
    a = create_body(Rect, 50, 50, pos=(0, 0))
    b = create_body(Circle, 5, pos=(0, 0))
    assert_collides(a, b)


def test_circle_hits_box_corner():
    # Box at origin (extents +/- 10).
    # Circle (r=5) at (12, 12).
    # Corner is at (10, 10). Dist to circle center is sqrt(2^2 + 2^2) = 2.82.
    # Radius is 5. 2.82 < 5. They collide.
    a = create_body(Rect, 20, 20, pos=(0, 0))
    b = create_body(Circle, 5, pos=(12, 12))
    assert_collides(a, b)


def test_circle_misses_box_corner():
    # Same setup but circle moved to (16, 16).
    # Dist to corner (10,10) is sqrt(6^2 + 6^2) = 8.48.
    # Radius is 5. 8.48 > 5. No collision.
    a = create_body(Rect, 20, 20, pos=(0, 0))
    b = create_body(Circle, 5, pos=(16, 16))
    assert_separate(a, b)



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Callable, Self
import bisect
import math
import random
import string
import io
import base64
import imageio
import cairosvg


from .base import IVisual, Text, Polyline
from .core import IShape, Point, Shape, Transform
from .canvas import Canvas


# --- Easing Functions ---
def linear(t: float) -> float:
    return t


def smooth(t: float) -> float:
    return t * t * (3 - 2 * t)


def ease_out(t: float) -> float:
    return t * (2 - t)


def ease_in_out_cubic(t: float) -> float:
    return 3 * t * t - 2 * t * t * t if t < 0.5 else 1 - (-2 * t + 2) ** 3 / 2


# --- Core Animation Class ---


class Animation(ABC):
    """
    Base class for logic that modifies a shape over a normalized time t [0, 1].
    """

    def __init__(self):
        self.rate = linear
        self._weight = 1.0
        self._started = False

    def begin(self) -> None:
        """Capture initial state. Only run once per playback."""
        if not self._started:
            self._start()
            self._started = True

    def _start(self) -> None:
        """Subclasses implement specific state capture here."""
        pass

    def update(self, t: float) -> None:
        """Apply changes based on normalized progress t."""
        if not self._started:
            raise TypeError("You need to call begin() first.")

        self._update(self.rate(t))

    @abstractmethod
    def _update(self, t: float):
        pass

    def finish(self) -> None:
        """Force final state."""
        if not self._started:
            self.begin()

        self.update(1.0)

    # --- Fluent API Modifiers ---

    def weight(self, w: float) -> Self:
        self._weight = w
        return self

    def rated(self, func: Callable[[float], float]) -> Self:
        self.rate = func
        return self

    def smoothed(self) -> Self:
        return self.rated(smooth)

    def delayed(self, delay_ratio: float) -> Animation:
        return Wrapped(self).rated(
            lambda t: 0 if t < delay_ratio else (t - delay_ratio) / (1 - delay_ratio),
        )

    def repeating(self, times: float) -> Animation:
        return Wrapped(self).rated(lambda t: (t * times) % 1.0)

    def looping(self) -> Animation:
        return Wrapped(self).rated(lambda t: 2 * t if t < 0.5 else 2 - 2 * t)

    def reversed(self) -> Animation:
        return Wrapped(self).rated(lambda t: 1.0 - t)

    # --- Operator Overloads ---

    def __mul__(self, factor: float) -> Animation:
        return self.repeating(factor)

    def __or__(self, other: Animation) -> Animation:
        children = self.children if isinstance(self, Sequence) else [self]

        if isinstance(other, Sequence):
            children.extend(other.children)
        else:
            children.append(other)

        return Sequence(*children)

    def __add__(self, other: Animation) -> Animation:
        children = self.children if isinstance(self, Parallel) else [self]

        if isinstance(other, Sequence):
            children.extend(other.children)
        else:
            children.append(other)

        return Parallel(*children)


# --- Structural Animations ---


class Wrapped(Animation):
    def __init__(self, child: Animation):
        super().__init__()
        self.child = child
        self.relative_weight = child._weight

    def _start(self):
        self.child.begin()

    def _update(self, t: float):
        self.child.update(t)


class Sequence(Animation):
    def __init__(self, *animations: Animation, **kwargs):
        super().__init__(**kwargs)
        self.children = list(animations)

        total_weight = sum(c._weight for c in animations)
        self.checkpoints = []
        current = 0.0

        for c in animations:
            width = c._weight / (total_weight or 1)
            current += width
            self.checkpoints.append(current)

    def _update(self, t: float):
        n = len(self.children)

        if n == 0:
            return

        idx = bisect.bisect_left(self.checkpoints, t)

        if idx >= n:
            idx = n - 1

        start_t = self.checkpoints[idx - 1] if idx > 0 else 0.0
        end_t = self.checkpoints[idx]
        duration = end_t - start_t

        local_t = (t - start_t) / duration if duration > 1e-9 else 1.0

        # Catch up previous
        for i in range(idx):
            child = self.children[i]

            if not child._started:
                child.begin()

            child.update(1.0)

        # Update active
        active = self.children[idx]

        if not active._started:
            active.begin()

        active.update(local_t)


class Parallel(Animation):
    def __init__(self, *animations: Animation, **kwargs):
        super().__init__(**kwargs)
        self.children = list(animations)

    def _start(self) -> None:
        for anim in self.children:
            anim.begin()

    def _update(self, t: float):
        for anim in self.children:
            anim.update(t)


class Delayed(Animation):
    """
    Plays a group of animations with a time lag.
    lag_ratio: 0.0 (parallel) -> 1.0 (sequential)
    """

    def __init__(self, *animations: Animation, lag_ratio: float = 0.1, **kwargs):
        super().__init__(**kwargs)
        self.anims = animations
        self.lag = lag_ratio

    def begin(self):
        self._started = True

    def update(self, t: float):
        t = self.rate(t)
        n = len(self.anims)
        if n == 0:
            return

        # Total span in relative time units = 1.0 + total lag
        total_span = 1.0 + (n - 1) * self.lag

        for i, anim in enumerate(self.anims):
            # Calculate start/end time for this specific animation in global t
            start = (i * self.lag) / total_span
            end = (1.0 + i * self.lag) / total_span

            if t < start:
                local_t = 0.0
            elif t > end:
                local_t = 1.0
            else:
                local_t = (t - start) / (end - start)

            if local_t > 0 and not anim._started:
                anim.begin()

            anim.update(local_t)


class Wait(Animation):
    def __init__(self, weight: float = 1.0):
        super().__init__()
        self.weight(weight)

    def _update(self, t: float):
        return super()._update(t)


# --- Property Animations ---


class Transformed(Animation):
    def __init__(self, shape: IShape, target: Transform, **kwargs):
        super().__init__(**kwargs)
        self.shape = shape
        self.target = target
        self.start_transform = None

    def _start(self):
        self.start_transform = self.shape.transform.copy()

    def _update(self, t: float):
        if self.start_transform is None:
            raise TypeError("You need to call begin() first.")

        new_trans = self.start_transform.lerp(self.target, t)
        self.shape.transform = new_trans


class Styled(Animation):
    def __init__(self, shape: IVisual, fill=None, stroke=None, width=None, **kwargs):
        super().__init__(**kwargs)
        self.shape = shape
        self.target_fill = fill
        self.target_stroke = stroke
        self.target_width = width
        self.start_fill = None
        self.start_stroke = None
        self.start_width = None

    def _start(self):
        self.start_fill = self.shape.fill
        self.start_stroke = self.shape.stroke
        self.start_width = self.shape.width

    def _update(self, t: float):

        if (
            self.start_fill is None
            or self.start_stroke is None
            or self.start_width is None
        ):
            raise TypeError("You need to call begin() first.")

        if self.target_width is not None:
            self.shape.width = (
                self.start_width + (self.target_width - self.start_width) * t
            )
        if self.target_fill is not None:
            self.shape.fill = self.start_fill.lerp(self.target_fill, t)
        if self.target_stroke is not None:
            self.shape.stroke = self.start_stroke.lerp(self.target_stroke, t)


# --- Specialized Animations ---


class Written(Animation):
    def __init__(self, shape: Text, **kwargs):
        super().__init__(**kwargs)
        self.shape = shape
        self.full_text = shape.content

    def _start(self):
        self.full_text = self.shape.content

    def _update(self, t: float):
        count = int(t * len(self.full_text))
        self.shape.content = self.full_text[:count]


class Scrambled(Animation):
    def __init__(self, shape: Text, seed: int = 42, **kwargs):
        super().__init__(**kwargs)
        self.shape = shape
        self.full_text = self.shape.content
        self.rng = random.Random(seed)

    def _start(self):
        self.full_text = self.shape.content

    def _update(self, t: float):
        total = len(self.full_text)
        resolved = int(t * total)

        result = list(self.full_text[:resolved])
        remaining = total - resolved
        chars = string.ascii_letters + string.digits + "!@#$%"
        scramble = [self.rng.choice(chars) for _ in range(remaining)]
        self.shape.content = "".join(result + scramble)


class Morphed(Animation):
    def __init__(self, shape: Polyline, target_points: list, **kwargs):
        super().__init__(**kwargs)
        self.shape = shape
        self.target_points = target_points
        self.start_points = []
        # Lazy import for Point to avoid circular dependency issues at module level
        from .core import Point

        self._Point = Point

    def _start(self):
        if not hasattr(self.shape, "points"):
            # Fail gracefully or raise
            self.start_points = []
            return

        current = getattr(self.shape, "points")
        self.start_points = [self._Point(p.x, p.y) for p in current]

        # Simple safety check on length
        if len(self.start_points) != len(self.target_points):
            print(
                f"Warning: Morph mismatch {len(self.start_points)} vs {len(self.target_points)}"
            )

    def _update(self, t: float):
        if not self.start_points:
            return

        new_pts = []
        for s, e in zip(self.start_points, self.target_points):
            nx = s.x + (e.x - s.x) * t
            ny = s.y + (e.y - s.y) * t
            new_pts.append(self._Point(nx, ny))

        self.shape.points = new_pts
        self.shape.refresh()


class Following(Animation):
    def __init__(
        self, shape: IShape, path: Shape, rotate_along: bool = False, **kwargs
    ):
        super().__init__(**kwargs)
        self.shape = shape
        self.path = path
        self.rotate_along = rotate_along

    def _update(self, t: float):
        # Requires path to have point_at(t) -> Point
        if not hasattr(self.path, "point_at"):
            return

        target_point = self.path.point_at(alpha)

        # Set position directly (Global coordinates)
        self.shape.transform.tx = target_point.x
        self.shape.transform.ty = target_point.y

        if self.rotate_along:
            epsilon = 0.01
            # Look ahead for tangent
            next_t = min(1.0, t + epsilon)
            # If we are at the end, look backward
            if next_t == t:
                next_t = t
                t = max(0.0, t - epsilon)
                target_point = self.path.point_at(t)

            next_p = self.path.point_at(next_t)
            angle = math.atan2(next_p.y - target_point.y, next_p.x - target_point.x)
            self.shape.transform.rotation = angle


class Warped(Animation):
    """
    Applies a time-dependent transformation function to a shape's points.
    The function should accept a Point and the current time t [0, 1].
    """

    def __init__(
        self, shape: Polyline, func: Callable[[Point, float], Point], **kwargs
    ):
        super().__init__(**kwargs)
        self.shape = shape
        self.func = func
        self.start_points: list[Point] = []

    def _start(self):
        # Capture the points exactly as they are when the animation begins
        if hasattr(self.shape, "points"):
            self.start_points = list(self.shape.points)

    def _update(self, t: float):
        if not self.start_points:
            return

        # Map the original points through the function at the current time t
        self.shape.points = [self.func(p, t) for p in self.start_points]

        # If the shape is a Component (like Polyline), it may need to rebuild its path
        if hasattr(self.shape, "refresh"):
            self.shape.refresh()


class NumericAnimation(Animation):
    """
    Animates any float/int attribute on an object.
    """

    def __init__(self, target: object, attribute: str, value: float, **kwargs):
        super().__init__(**kwargs)
        self.target_obj = target
        self.attribute = attribute
        self.value = value
        self.start = 0.0

    def _start(self):
        # Capture the current value when animation begins
        try:
            val = getattr(self.target_obj, self.attribute)
            self.start = float(val)
        except AttributeError:
            raise AttributeError(
                f"Object {self.target_obj} has no attribute '{self.attribute}'"
            )

    def _update(self, t: float):
        # Lerp: start + (end - start) * t
        current = self.start + (self.value - self.start) * t

        # Apply using setter
        setattr(self.target_obj, self.attribute, current)


class FunctionalAnimation(Animation):
    def __init__(self, obj: IShape, func: Callable):
        super().__init__()
        self.func = func
        self.obj = obj

    def _update(self, t: float):
        self.func(self.obj, t)


class KeyframeAnimation(Animation):
    """
    Animates multiple properties using keyframes.
    Usage: KeyframeAnimation(shape, tx={1.0: 100}, rotation={0.5: deg(180), 1.0: deg(360)})
    """

    def __init__(self, shape: IShape, **tracks):
        super().__init__()
        self.shape = shape
        self.raw_tracks = tracks
        self.compiled_tracks = {}

    def _start(self):
        self.compiled_tracks = {}

        for name, keyframes in self.raw_tracks.items():
            # 1. Resolve Target (Transform vs Shape)
            # Priority: Transform property -> Shape property
            target = self.shape
            if hasattr(self.shape, "transform") and hasattr(self.shape.transform, name):
                target = self.shape.transform
            elif not hasattr(self.shape, name):
                raise AttributeError(
                    f"Property '{name}' not found on shape or transform."
                )

            # 2. Capture Current Value
            current_val = getattr(target, name)

            # 3. Sort and Normalize Frames
            # Structure: [(time, value, easing_func), ...]
            sorted_times = sorted(keyframes.keys())
            frames = []

            # Automatic 0.0 Keyframe
            if 0.0 not in keyframes:
                frames.append((0.0, current_val, linear))

            for t in sorted_times:
                entry = keyframes[t]

                # Support tuple for (value, easing)
                if isinstance(entry, tuple):
                    val, func = entry
                else:
                    val, func = entry, linear

                frames.append((t, val, func))

            # Separate times for bisect
            times = [f[0] for f in frames]
            self.compiled_tracks[name] = (target, times, frames)

    def _update(self, t: float):
        for name, (target, times, frames) in self.compiled_tracks.items():
            # Find the interval [t0, t1] where t0 <= t <= t1
            idx = bisect.bisect_right(times, t)

            # Clamp to boundaries
            if idx == 0:
                setattr(target, name, frames[0][1])
                continue
            if idx >= len(frames):
                setattr(target, name, frames[-1][1])
                continue

            # Get surrounding frames
            t0, v0, _ = frames[idx - 1]
            t1, v1, easing = frames[idx]

            # Normalize t to the segment 0..1
            segment_duration = t1 - t0
            if segment_duration <= 1e-9:
                local_t = 1.0
            else:
                local_t = (t - t0) / segment_duration

            # Apply segment easing
            eased_t = easing(local_t)

            # Interpolate
            current = v0 + (v1 - v0) * eased_t
            setattr(target, name, current)


# --- Factory Class ---


class Animator[TShape: IShape]:
    """Helper attached to shape.animate"""

    def __init__(self, shape: TShape):
        self.shape = shape

    # Transform
    def translate(self, dx: float = 0, dy: float = 0) -> Animation:
        target = self.shape.transform.copy()
        target.tx += dx
        target.ty += dy
        return Transformed(self.shape, target)

    def rotate(self, angle: float) -> Animation:
        target = self.shape.transform.copy()
        target.rotation += angle
        return Transformed(self.shape, target)

    def scale(self, factor: float) -> Animation:
        target = self.shape.transform.copy()
        target.sx *= factor
        target.sy *= factor
        return Transformed(self.shape, target)

    def follow(self, path: Shape, rotate: bool = False) -> Animation:
        return Following(self.shape, path, rotate_along=rotate)

    def property(self, name: str, value: float) -> Animation:
        """Explicitly animate a property by name."""
        return NumericAnimation(self.shape, name, value)

    def custom(self, func: Callable[[TShape, float]]) -> Animation:
        """Custom, function-based animation."""
        return FunctionalAnimation(self.shape, func)

    def __getattr__(self, name: str):
        """
        Magic method: shape.animate.foo(100) -> animates 'foo' to 100.
        Returns a callable that creates the animation.
        """

        def proxy(target: float):
            return NumericAnimation(self.shape, name, target)

        return proxy

    def keyframes(self, **tracks) -> Animation:
        return KeyframeAnimation(self.shape, **tracks)


class StyledAnimator[TShape: IVisual](Animator[TShape]):
    def __init__(self, shape: TShape):
        super().__init__(shape)

    # Style
    def fill(self, color: str) -> Animation:
        return Styled(self.shape, fill=color)

    def stroke(self, color: str) -> Animation:
        return Styled(self.shape, stroke=color)

    def style(self, **kwargs) -> Animation:
        return Styled(self.shape, **kwargs)


class TextAnimator(StyledAnimator[Text]):
    def __init__(self, shape: Text):
        super().__init__(shape)

    # Text
    def write(self) -> Animation:
        return Written(self.shape)

    def scramble(self) -> Animation:
        return Scrambled(self.shape)


class PolylineAnimator(StyledAnimator[Polyline]):
    # Geometry
    def morph(self, target: Polyline) -> Animation:
        pts = target.points
        return Morphed(self.shape, pts)

    def warp(self, func: Callable[[Point, float], Point]) -> Animation:
        """
        Creates a warp animation using a function that maps (Point, t) -> Point.
        """
        return Warped(self.shape, func)


# --- Scene Class ---


class Scene:
    def __init__(self, canvas: Canvas, fps: int = 30, background: str = "white"):
        self.canvas = canvas
        self.fps = fps
        self.background = background
        self._frames: list[bytes] = []

    def capture(self):
        svg = self.canvas._build_svg()
        png = cairosvg.svg2png(
            bytestring=svg.encode("utf-8"), background_color=self.background
        )
        self._frames.append(png)

    def play(self, *animations: Animation, duration: float = 1.0):
        if not animations:
            return

        total = int(duration * self.fps)

        # Initialize
        for anim in animations:
            anim.begin()

        # Loop
        for i in range(total):
            t = i / total

            for anim in animations:
                anim.update(t)

            self.capture()

        # Finalize
        for anim in animations:
            anim.finish()

    def wait(self, duration: float):
        self.play(Wait(), duration=duration)

    def save(self, dest, format=None):
        if not self._frames:
            return
        images = [imageio.imread(io.BytesIO(f)) for f in self._frames]

        kwargs = {"fps": self.fps}
        if format == "mp4":
            kwargs["quality"] = 8
        else:
            kwargs["loop"] = 0

        imageio.mimsave(dest, images, format=format or "gif", **kwargs)

    def display(self):
        try:
            from IPython.display import display, HTML
        except ImportError:
            return

        buf = io.BytesIO()
        self.save(buf, format="gif")
        b64 = base64.b64encode(buf.getvalue()).decode("utf-8")
        display(
            HTML(f'<img src="data:image/gif;base64,{b64}" style="max-width:100%"/>')
        )

    def _ipython_display_(self):
        self.display()



from dataclasses import dataclass


@dataclass
class Collider:
    pass


@dataclass
class CircleCollider(Collider):
    radius: float


@dataclass
class BoxCollider(Collider):
    width: float
    height: float



from __future__ import annotations
import math
from dataclasses import dataclass
from typing import Callable, cast
from .core import Body, Point
from .colliders import CircleCollider, BoxCollider, Collider

SOLVERS = dict()


def solver[C1: Collider, C2: Collider](t1: type[C1], t2: type[C2]):
    def wrapper(func: Callable[[Body, Body], Collision | None]):
        Collision.register(t1, t2, func)
        return func

    return wrapper


@dataclass
class Collision:
    """Stores information about a collision between A and B."""

    a: Body
    b: Body
    normal: Point  # A -> B
    depth: float
    point: Point  # Contact Point in World Space

    @staticmethod
    def register(t1, t2, func):
        SOLVERS[(t1, t2)] = func
        SOLVERS[(t2, t1)] = lambda a, b: func(b, a)._flip() if func(b, a) else None

    def _flip(self) -> Collision:
        self.normal = self.normal * -1
        self.a, self.b = self.b, self.a
        return self

    @classmethod
    def solve(cls, a: Body, b: Body) -> Collision | None:
        t1, t2 = type(a.collider), type(b.collider)
        if s := SOLVERS.get((t1, t2)):
            return s(a, b)
        return None

    def resolve(self):
        """Applies Rigid Body Impulse with Friction."""
        a, b = self.a, self.b
        n = self.normal

        # 1. Positional Correction (Anti-Sinking)
        total_inv_mass = a.inv_mass + b.inv_mass
        if total_inv_mass == 0:
            return

        # Increased from 0.5 to 0.8 to reduce "mushiness"
        correction = max(self.depth - 0.01, 0.0) / total_inv_mass * 0.8
        move = n * correction
        if not a.static:
            a.pos -= move * a.inv_mass
        if not b.static:
            b.pos += move * b.inv_mass

        # 2. Lever Arms (Vector from Center to Contact Point)
        ra = self.point - a.pos
        rb = self.point - b.pos

        # 3. Relative Velocity at Contact Point
        # V_p = V_cm + (w x r)
        # 2D Cross scalar-vector: w x r = (-w*ry, w*rx)
        va = a.vel + Point(-a.angular_vel * ra.y, a.angular_vel * ra.x)
        vb = b.vel + Point(-b.angular_vel * rb.y, b.angular_vel * rb.x)

        rv = vb - va

        # Velocity along normal
        vel_along_normal = rv.x * n.x + rv.y * n.y
        if vel_along_normal > 0:
            return  # Separating

        # 4. Compute Impulse Scalar (J)
        raxn = ra.x * n.y - ra.y * n.x  # Cross product 2D (scalar)
        rbxn = rb.x * n.y - rb.y * n.x

        inv_mass_sum = (
            a.inv_mass
            + b.inv_mass
            + (raxn**2 * a.inv_inertia)
            + (rbxn**2 * b.inv_inertia)
        )

        e = min(a.material.restitution, b.material.restitution)
        j = -(1 + e) * vel_along_normal
        j /= inv_mass_sum

        impulse = n * j

        self._apply_impulse(impulse, ra, rb)

        # 5. Friction Impulse
        t = (rv - (n * vel_along_normal)).normalize()

        raxt = ra.x * t.y - ra.y * t.x
        rbxt = rb.x * t.y - rb.y * t.x
        inv_mass_sum_t = (
            a.inv_mass
            + b.inv_mass
            + (raxt**2 * a.inv_inertia)
            + (rbxt**2 * b.inv_inertia)
        )

        jt = -(rv.x * t.x + rv.y * t.y)
        jt /= inv_mass_sum_t

        mu = math.sqrt(a.material.friction * b.material.friction)

        max_j = j * mu
        if abs(jt) > max_j:
            friction_impulse = t * (max_j * (1 if jt > 0 else -1))
        else:
            friction_impulse = t * jt

        self._apply_impulse(friction_impulse, ra, rb)

    def _apply_impulse(self, impulse: Point, ra: Point, rb: Point):
        a, b = self.a, self.b

        if not a.static:
            a.vel -= impulse * a.inv_mass
        if not b.static:
            b.vel += impulse * b.inv_mass

        torque_a = ra.x * impulse.y - ra.y * impulse.x
        torque_b = rb.x * impulse.y - rb.y * impulse.x

        if not a.static:
            a.angular_vel -= torque_a * a.inv_inertia
        if not b.static:
            b.angular_vel += torque_b * b.inv_inertia


# --- Helpers ---


def _get_box_vertices(b: Body) -> list[Point]:
    box = cast(BoxCollider, b.collider)
    hw, hh = box.width / 2, box.height / 2
    corners = [Point(-hw, -hh), Point(hw, -hh), Point(hw, hh), Point(-hw, hh)]

    rot = getattr(b, "rotation", 0.0)
    pos = b.pos
    cos_r, sin_r = math.cos(rot), math.sin(rot)

    world_corners = []
    for p in corners:
        rx = p.x * cos_r - p.y * sin_r
        ry = p.x * sin_r + p.y * cos_r
        world_corners.append(Point(rx + pos.x, ry + pos.y))
    return world_corners


def _get_axes(verts: list[Point]) -> list[Point]:
    axes = []
    for i in range(len(verts)):
        p1 = verts[i]
        p2 = verts[(i + 1) % len(verts)]
        edge = p2 - p1
        axes.append(Point(-edge.y, edge.x).normalize())
    return axes


def _project(verts: list[Point], axis: Point) -> tuple[float, float]:
    min_p = float("inf")
    max_p = float("-inf")
    for v in verts:
        proj = v.x * axis.x + v.y * axis.y
        if proj < min_p:
            min_p = proj
        if proj > max_p:
            max_p = proj
    return min_p, max_p


# --- Solvers ---


@solver(CircleCollider, CircleCollider)
def circle_to_circle(a: Body, b: Body) -> Collision | None:
    n = b.pos - a.pos
    dist = n.magnitude()
    r = (
        cast(CircleCollider, a.collider).radius
        + cast(CircleCollider, b.collider).radius
    )

    if dist > r:
        return None

    normal = n.normalize() if dist > 0 else Point(1, 0)
    contact = a.pos + (normal * cast(CircleCollider, a.collider).radius)

    return Collision(a, b, normal, r - dist, contact)


@solver(CircleCollider, BoxCollider)
def circle_to_box(circ: Body, box: Body) -> Collision | None:
    # 1. Transform Circle to Box Local
    box_rot = getattr(box, "rotation", 0.0)
    rel = circ.pos - box.pos
    cos_r, sin_r = math.cos(-box_rot), math.sin(-box_rot)
    lx = rel.x * cos_r - rel.y * sin_r
    ly = rel.x * sin_r + rel.y * cos_r

    hw = cast(BoxCollider, box.collider).width / 2
    hh = cast(BoxCollider, box.collider).height / 2

    # 2. Closest Point on Box (Local)
    cx = max(-hw, min(lx, hw))
    cy = max(-hh, min(ly, hh))

    dx, dy = lx - cx, ly - cy
    dist_sq = dx**2 + dy**2
    r = cast(CircleCollider, circ.collider).radius

    if dist_sq > r**2:
        return None

    dist = math.sqrt(dist_sq)
    if dist == 0:
        local_n = Point(0, -1)
        depth = r
    else:
        local_n = Point(dx / dist, dy / dist)
        depth = r - dist

    # Rotate Normal to World
    # Note: local_n points Box->Circle. This matches A->B convention?
    # No, A=Circle, B=Box. We want A->B. Current is B->A.
    local_n = local_n * -1

    cos_w, sin_w = math.cos(box_rot), math.sin(box_rot)
    nx = local_n.x * cos_w - local_n.y * sin_w
    ny = local_n.x * sin_w + local_n.y * cos_w

    wx = cx * cos_w - cy * sin_w
    wy = cx * sin_w + cy * cos_w
    contact = box.pos + Point(wx, wy)

    return Collision(circ, box, Point(nx, ny), depth, contact)


@solver(BoxCollider, BoxCollider)
def box_to_box(a: Body, b: Body) -> Collision | None:
    verts_a = _get_box_vertices(a)
    verts_b = _get_box_vertices(b)

    axes_a = _get_axes(verts_a)
    axes_b = _get_axes(verts_b)
    all_axes = axes_a + axes_b

    min_overlap = float("inf")
    best_axis = Point(0, 0)

    # 1. SAT Check
    for axis in all_axes:
        min_a, max_a = _project(verts_a, axis)
        min_b, max_b = _project(verts_b, axis)

        if max_a < min_b or max_b < min_a:
            return None

        overlap = min(max_a, max_b) - max(min_a, min_b)
        if overlap < min_overlap:
            min_overlap = overlap
            best_axis = axis

    # 2. Correct Normal A -> B
    direction = b.pos - a.pos
    if direction.x * best_axis.x + direction.y * best_axis.y < 0:
        best_axis = best_axis * -1

    # 3. Identify Contact Point (Reference Face vs Incident Vertex)
    # The body with a face most perpendicular to the normal is the Reference.
    def get_max_axis_dot(axes, n):
        max_d = 0.0
        for ax in axes:
            d = abs(ax.x * n.x + ax.y * n.y)
            if d > max_d:
                max_d = d
        return max_d

    dot_a = get_max_axis_dot(axes_a, best_axis)
    dot_b = get_max_axis_dot(axes_b, best_axis)

    if dot_a > dot_b:
        # A is Reference (Flat Face), B is Incident (Corner)
        # Find B's vertex most opposite to Normal (deepest into A)
        candidates = verts_b
        best_v = candidates[0]
        min_d = float("inf")
        for v in candidates:
            d = v.x * best_axis.x + v.y * best_axis.y
            if d < min_d:
                min_d = d
                best_v = v
        contact = best_v
    else:
        # B is Reference (Flat Face), A is Incident (Corner)
        # Find A's vertex most along Normal (deepest into B)
        candidates = verts_a
        best_v = candidates[0]
        max_d = float("-inf")
        for v in candidates:
            d = v.x * best_axis.x + v.y * best_axis.y
            if d > max_d:
                max_d = d
                best_v = v
        contact = best_v

    return Collision(a, b, best_axis, min_overlap, contact)



import math
import pytest
from tesserax.core import Point, Bounds
from tesserax.base import Rect, Circle, Group

# --- Fixtures ---


@pytest.fixture
def messy_group():
    """Returns a group of shapes scattered randomly."""
    g = Group()
    g.add(Rect(10, 10).translated(0, 0))
    g.add(Rect(20, 20).translated(50, 50))  # Far away
    g.add(Rect(15, 15).translated(-20, 10))  # Overlapping
    return g


@pytest.fixture
def sizes_group():
    """Group with varied sizes to test alignment robustness."""
    g = Group()
    g.add(Rect(10, 50))  # Tall
    g.add(Rect(10, 10))  # Small
    g.add(Rect(10, 30))  # Medium
    return g


# --- Alignment Invariants ---


def test_top_alignment_invariant(sizes_group: Group):
    """Invariant: After aligning top, all min_y values must be equal."""
    # Pre-condition: y values are different (implied by centered construction with different heights)

    sizes_group.align(axis="vertical", anchor="top")

    # Capture the "y" of the first element
    expected_y = sizes_group.shapes[0].bounds().y

    for s in sizes_group.shapes:
        # All shapes must match this Y
        assert math.isclose(s.bounds().y, expected_y, abs_tol=1e-9)


def test_center_alignment_invariant(sizes_group: Group):
    """Invariant: After aligning centers, midpoints must match."""
    sizes_group.align(axis="both", anchor="center")

    first_center = sizes_group.shapes[0].bounds().center

    for s in sizes_group.shapes:
        c = s.bounds().center
        assert math.isclose(c.x, first_center.x, abs_tol=1e-9)
        assert math.isclose(c.y, first_center.y, abs_tol=1e-9)


# --- Distribution Invariants ---


def test_horizontal_monotonicity(messy_group: Group):
    """Invariant: In a distributed row, Right(i) <= Left(i+1)."""
    gap = 5.0
    messy_group.distribute(axis="horizontal", gap=gap)

    shapes = messy_group.shapes
    for i in range(len(shapes) - 1):
        current = shapes[i].bounds()
        next_shape = shapes[i + 1].bounds()

        # The right edge of current + gap must equal left edge of next
        # (Use isclose because floating point math)
        assert math.isclose(current.right.x + gap, next_shape.x, abs_tol=1e-5)


def test_vertical_monotonicity(messy_group: Group):
    """Invariant: In a distributed col, Bottom(i) <= Top(i+1)."""
    gap = 10.0
    messy_group.distribute(axis="vertical", gap=gap)

    shapes = messy_group.shapes
    for i in range(len(shapes) - 1):
        current = shapes[i].bounds()
        next_shape = shapes[i + 1].bounds()

        assert math.isclose(current.bottom.y + gap, next_shape.y, abs_tol=1e-5)


def test_gap_integrity():
    """Invariant: Space between items exactly matches requested gap."""
    g = Group()
    r1 = Rect(10, 10)
    r2 = Rect(10, 10)
    g.add(r1, r2)

    target_gap = 12.345
    g.distribute(axis="horizontal", gap=target_gap)

    distance = r2.bounds().x - r1.bounds().right.x
    assert math.isclose(distance, target_gap, abs_tol=1e-9)


# --- Nesting Invariants ---


def test_group_translation_preserves_relative_layout(sizes_group: Group):
    """Invariant: Moving a Group moves its children but maintains their relative positions."""
    sizes_group.distribute(axis="horizontal", gap=5)

    # Snapshot relative distance between 0 and 1
    dist_before = sizes_group.shapes[1].bounds().x - sizes_group.shapes[0].bounds().x

    # Move the entire group massively
    sizes_group.translated(1000, 500)

    # Relative distance must be identical
    dist_after = sizes_group.shapes[1].bounds().x - sizes_group.shapes[0].bounds().x
    assert math.isclose(dist_before, dist_after, abs_tol=1e-9)



from .canvas import Canvas
from .core import Shape, Bounds, Point, deg
from .base import (
    Rect,
    Square,
    Circle,
    Ellipse,
    Line,
    Arrow,
    Group,
    Path,
    Polyline,
    Text,
)

__version__ = "0.8.1"



from __future__ import annotations
import math
from typing import Callable, Literal, Protocol, Self, TYPE_CHECKING, cast

from tesserax.color import Color, Colors
from .core import Anchor, IShape, Point, Shape, Bounds, Component

if TYPE_CHECKING:
    from .animation import StyledAnimator, TextAnimator, PolylineAnimator


class IVisual(IShape):
    fill: Color
    stroke: Color
    width: float


class Visual(Shape, IVisual):
    def __init__(
        self,
        fill: Color = Colors.Transparent,
        stroke: Color = Colors.Black,
        width: float = 1.0,
    ) -> None:
        super().__init__()
        self.fill = fill
        self.stroke = stroke
        self.width = width

    @property
    def animate(self) -> StyledAnimator:
        from .animation import StyledAnimator

        if self._animator is None:
            self._animator = StyledAnimator(self)

        return cast(StyledAnimator, self._animator)


class Rect(Visual):
    """A rectangular shape, the foundation for arrays and memory blocks."""

    def __init__(
        self,
        w: float,
        h: float,
        fill: Color = Colors.Transparent,
        stroke: Color = Colors.Black,
        width: float = 1.0,
    ) -> None:
        super().__init__(fill=fill, stroke=stroke, width=width)
        self.w, self.h = w, h

    def local(self) -> Bounds:
        # Returns bounds centered at 0,0
        hw, hh = self.w / 2, self.h / 2
        return Bounds(-hw, -hh, self.w, self.h)

    def _render(self) -> str:
        # Render centered at 0,0
        # x and y are top-left coordinates relative to the origin
        return f'<rect x="{-self.w/2}" y="{-self.h/2}" width="{self.w}" height="{self.h}" stroke="{self.stroke}" fill="{self.fill}" />'


class Square(Visual):
    """A specialized Rect where width equals height."""

    def __init__(
        self,
        size: float,
        fill: Color = Colors.Transparent,
        stroke: Color = Colors.Black,
        width: float = 1.0,
    ) -> None:
        super().__init__(fill=fill, stroke=stroke, width=width)
        self.size = size

    def local(self) -> Bounds:
        # Returns bounds centered at 0,0
        hw, hh = self.size / 2, self.size / 2
        return Bounds(-hw, -hh, self.size, self.size)

    def _render(self) -> str:
        # Render centered at 0,0
        # x and y are top-left coordinates relative to the origin
        return f'<rect x="{-self.size/2}" y="{-self.size/2}" width="{self.size}" height="{self.size}" stroke="{self.stroke}" fill="{self.fill}" />'


class Circle(Visual):
    """A circle, ideal for nodes in trees or states in automata."""

    def __init__(
        self,
        r: float,
        fill: Color = Colors.Transparent,
        stroke: Color = Colors.Black,
        width: float = 1.0,
    ) -> None:
        super().__init__(fill=fill, stroke=stroke, width=width)
        self.r = r

    def local(self) -> Bounds:
        return Bounds(-self.r, -self.r, self.r * 2, self.r * 2)

    def _render(self) -> str:
        # cx, cy are center offsets relative to local transform (which is 0,0)
        return f'<circle cx="0" cy="0" r="{self.r}" stroke="{self.stroke}" fill="{self.fill}" />'


class Ellipse(Visual):
    """An ellipse for when text labels are wider than they are tall."""

    def __init__(
        self,
        rx: float,
        ry: float,
        fill: Color = Colors.Transparent,
        stroke: Color = Colors.Black,
        width: float = 1.0,
    ) -> None:
        super().__init__(fill=fill, stroke=stroke, width=width)
        self.rx, self.ry = rx, ry

    def local(self) -> Bounds:
        return Bounds(-self.rx, -self.ry, self.rx * 2, self.ry * 2)

    def _render(self) -> str:
        return f'<ellipse cx="0" cy="0" rx="{self.rx}" ry="{self.ry}" stroke="{self.stroke}" fill="{self.fill}" />'


class Group(Shape):
    stack: list[list[Shape]] = []

    @classmethod
    def current(cls) -> list[Shape] | None:
        if cls.stack:
            return cls.stack[-1]
        return None

    """A collection of shapes that behaves as a single unit."""

    def __init__(self, shapes: list[Shape] | None = None) -> None:
        super().__init__()
        self.shapes: list[Shape] = []

        if shapes:
            self.add(*shapes)

    def add(self, *shapes: Shape, mode: Literal["strict", "loose"] = "strict") -> Group:
        """Adds a shape and returns self for chaining."""
        for shape in shapes:
            if shape.parent:
                if mode == "strict":
                    raise RuntimeError("This shape already has a parent")
                else:
                    shape.parent.remove(shape)

            self.shapes.append(shape)
            shape.parent = self

        return self

    def remove(self, shape: Shape) -> Self:
        self.shapes.remove(shape)
        shape.parent = None
        return self

    def local(self) -> Bounds:
        """Computes the union of all child bounds."""
        if not self.shapes:
            return Bounds(0, 0, 0, 0)

        return Bounds.union(*[s.bounds() for s in self.shapes])

    def _render(self) -> str:
        return "\n".join(s.render() for s in self.shapes)

    def __iadd__(self, other: Shape) -> Self:
        """Enables 'group += shape'."""
        self.shapes.append(other)
        return self

    def __enter__(self):
        self.stack.append([])
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.add(*self.stack.pop(), mode="loose")

    def align(
        self,
        axis: Literal["horizontal", "vertical", "both"] = "both",
        anchor: Anchor = "center",
    ) -> Self:
        """
        Aligns all children in the group relative to the anchor of the first child.
        """
        if not self.shapes:
            return self

        first = self.shapes[0]
        # Map local anchor to parent space for comparison
        ref_p = first.transform.map(first.local().anchor(anchor))

        for shape in self.shapes[1:]:
            curr_p = shape.transform.map(shape.local().anchor(anchor))

            if axis in ("horizontal", "both"):
                shape.transform.tx += ref_p.x - curr_p.x

            if axis in ("vertical", "both"):
                shape.transform.ty += ref_p.y - curr_p.y

        return self

    def distribute(
        self,
        axis: Literal["horizontal", "vertical"],
        size: float | None = None,
        mode: Literal["tight", "space-between", "space-around"] = "tight",
        gap: float = 0.0,
    ) -> Self:
        """
        Distributes children along an axis using rigid or flexible spacing.
        """
        if not self.shapes:
            return self

        for s in self.shapes:
            if isinstance(s, Spring):
                s._size = 0.0

        springs = [s for s in self.shapes if isinstance(s, Spring)]
        total_flex = sum(s.flex for s in springs)
        n = len(self.shapes)

        is_horiz = axis == "horizontal"
        # Calculate rigid total based on children's bounds in parent space?
        # Ideally we use local size, but shapes might be scaled.
        # For now, we use local().width/height assuming no rotation/scale on children yet.
        rigid_total = sum(
            s.local().width if is_horiz else s.local().height
            for s in self.shapes
            if not isinstance(s, Spring)
        )

        effective_gap = gap
        start_offset = 0.0
        spring_unit = 0.0

        if size is not None:
            if springs:
                remaining = size - rigid_total - (gap * (n - 1))
                spring_unit = max(0, remaining / total_flex) if total_flex > 0 else 0
            elif mode == "space-between" and n > 1:
                effective_gap = (size - rigid_total) / (n - 1)
            elif mode == "space-around":
                effective_gap = (size - rigid_total) / n
                start_offset = effective_gap / 2

        cursor = start_offset
        for s in self.shapes:
            b = s.local()

            if isinstance(s, Spring):
                s._size = s.flex * spring_unit
                cursor += s._size
            else:
                if is_horiz:
                    s.transform.tx = cursor - b.x
                    cursor += b.width + effective_gap
                else:
                    s.transform.ty = cursor - b.y
                    cursor += b.height + effective_gap

        return self


class Text(Visual):
    """
    Text primitive. Default anchor is 'middle' for true centering.
    """

    def __init__(
        self,
        content: str,
        size: float = 12,
        font: str = "sans-serif",
        anchor: Literal["start", "middle", "end"] = "middle",
        baseline: Literal["top", "middle", "bottom"] = "middle",
        fill: Color = Colors.Black,
        stroke: Color = Colors.Transparent,
        width: float = 1.0,
    ) -> None:
        super().__init__(fill=fill, stroke=stroke, width=width)
        self.content = content
        self.size = size
        self.font = font
        self._anchor = anchor
        self._baseline = baseline

    @property
    def animate(self) -> TextAnimator:
        from .animation import TextAnimator

        if self._animator is None:
            self._animator = TextAnimator(self)

        return cast(TextAnimator, self._animator)

    def local(self) -> Bounds:
        # Heuristic: average character width is ~60% of font size
        width = len(self.content) * self.size * 0.6
        height = self.size

        # Bounds logic for "middle" anchor (which is the default now)
        # origin (0,0) is at the center of the text.
        if self._anchor == "middle":
            return Bounds(-width / 2, -height / 2, width, height)
        elif self._anchor == "start":
            return Bounds(0, -height / 2, width, height)
        elif self._anchor == "end":
            return Bounds(-width, -height / 2, width, height)
        return Bounds(0, 0, width, height)

    def _render(self) -> str:
        # Note: 'dominant-baseline="middle"' centers text vertically around y=0
        return (
            f'<text x="0" y="0" font-family="{self.font}" font-size="{self.size}" '
            f'fill="{self.fill}" text-anchor="{self._anchor}" dominant-baseline="{self._baseline}">'
            f"{self.content}</text>"
        )


class Invisible(Shape):
    def _render(self) -> str:
        return ""


class Spacer(Invisible):
    def __init__(self, w: float, h: float) -> None:
        super().__init__()
        self.w, self.h = w, h

    def local(self) -> Bounds:
        return Bounds(0, 0, self.w, self.h)


class Ghost(Invisible):
    def __init__(self, target: Shape) -> None:
        super().__init__()
        self.target = target

    def local(self) -> Bounds:
        return self.target.local()


class Spring(Invisible):
    def __init__(self, flex: float = 1.0) -> None:
        super().__init__()
        self.flex = flex
        self._size: float = 0.0

    def local(self) -> Bounds:
        return Bounds(0, 0, self._size, self._size)


class Path(Visual):
    """
    A shape defined by an SVG path data string.
    """

    def __init__(
        self,
        fill: Color = Colors.Transparent,
        stroke: Color = Colors.Black,
        width: float = 1.0,
        marker_start: str | None = None,
        marker_end: str | None = None,
    ) -> None:
        super().__init__(fill=fill, stroke=stroke, width=width)
        self.marker_start = marker_start
        self.marker_end = marker_end
        self._reset()

    def _reset(self):
        self._commands: list[str] = []
        self._cursor: tuple[float, float] = (0.0, 0.0)
        self._min_x: float = float("inf")
        self._min_y: float = float("inf")
        self._max_x: float = float("-inf")
        self._max_y: float = float("-inf")

    def local(self) -> Bounds:
        if not self._commands:
            return Bounds(0, 0, 0, 0)
        width = self._max_x - self._min_x
        height = self._max_y - self._min_y
        return Bounds(self._min_x, self._min_y, width, height)

    def jump_to(self, x: float, y: float) -> Self:
        self._commands.append(f"M {x} {y}")
        self._update_cursor(x, y)
        return self

    def arc(
        self,
        rx: float,
        ry: float,
        rot: float,
        large: int,
        sweep: int,
        x: float,
        y: float,
    ) -> Self:
        self._commands.append(f"A {rx} {ry} {rot} {large} {sweep} {x} {y}")
        self._update_cursor(x, y)
        return self

    def jump_by(self, dx: float, dy: float) -> Self:
        x, y = self._cursor
        return self.jump_to(x + dx, y + dy)

    def line_to(self, x: float, y: float) -> Self:
        self._commands.append(f"L {x} {y}")
        self._update_cursor(x, y)
        return self

    def line_by(self, dx: float, dy: float) -> Self:
        x, y = self._cursor
        return self.line_to(x + dx, y + dy)

    def cubic_to(
        self,
        cp1_x: float,
        cp1_y: float,
        cp2_x: float,
        cp2_y: float,
        end_x: float,
        end_y: float,
    ) -> Self:
        self._commands.append(f"C {cp1_x} {cp1_y}, {cp2_x} {cp2_y}, {end_x} {end_y}")
        self._expand_bounds(cp1_x, cp1_y)
        self._expand_bounds(cp2_x, cp2_y)
        self._update_cursor(end_x, end_y)
        return self

    def quadratic_to(self, cx: float, cy: float, ex: float, ey: float) -> Self:
        self._commands.append(f"Q {cx} {cy}, {ex} {ey}")
        self._expand_bounds(cx, cy)
        self._update_cursor(ex, ey)
        return self

    def close(self) -> Self:
        self._commands.append("Z")
        return self

    def _update_cursor(self, x: float, y: float) -> None:
        self._cursor = (x, y)
        self._expand_bounds(x, y)

    def _expand_bounds(self, x: float, y: float) -> None:
        self._min_x = min(self._min_x, x)
        self._min_y = min(self._min_y, y)
        self._max_x = max(self._max_x, x)
        self._max_y = max(self._max_y, y)

    def _render(self) -> str:
        d = " ".join(self._commands)
        ms = f'marker-start="url(#{self.marker_start})"' if self.marker_start else ""
        me = f'marker-end="url(#{self.marker_end})"' if self.marker_end else ""
        return (
            f'<path d="{d}" fill="{self.fill}" '
            f'stroke="{self.stroke}" stroke-width="{self.width}" '
            f"{ms} {me} />"
        )


# Path-based components start here


class Polyline(Component, IVisual):
    """
    A sequence of connected lines.
    Supports .center() to re-align points around the origin.
    """

    def __init__(
        self,
        points: list[Point],
        smoothness: float = 0.0,
        closed: bool = False,
        fill: Color = Colors.Transparent,
        stroke: Color = Colors.Black,
        width: float = 1.0,
        marker_start: str | None = None,
        marker_end: str | None = None,
    ) -> None:
        super().__init__(
            fill=fill,
            stroke=stroke,
            width=width,
            marker_start=marker_start,
            marker_end=marker_end,
        )
        self.points = points or []
        self.smoothness = smoothness
        self.closed = closed

    @property
    def animate(self) -> PolylineAnimator:
        from .animation import PolylineAnimator

        if self._animator is None:
            self._animator = PolylineAnimator(self)

        return cast(PolylineAnimator, self._animator)

    def append(self, p: Point) -> Self:
        self.points.append(p)
        return self.refresh()

    def prepend(self, p: Point) -> Self:
        self.points.insert(0, p)
        return self.refresh()

    def extend(self, points: list[Point]) -> Self:
        self.points.extend(points)
        return self.refresh()

    def center(self) -> Self:
        """
        Shifts all points so their bounding box center is at (0,0).
        Then updates the transform translation to compensate,
        keeping the shape visually in the same place.
        """
        if not self.points:
            return self

        # 1. Measure current bounds
        min_x = min(p.x for p in self.points)
        max_x = max(p.x for p in self.points)
        min_y = min(p.y for p in self.points)
        max_y = max(p.y for p in self.points)

        cx = (min_x + max_x) / 2
        cy = (min_y + max_y) / 2

        # 2. Shift points to center
        new_points = []
        for p in self.points:
            new_points.append(Point(p.x - cx, p.y - cy))
        self.points = new_points

        # 3. Compensate transform
        self.transform.tx += cx
        self.transform.ty += cy

        # 4. Rebuild path
        return self

    def subdivide(self, times: int = 1) -> Self:
        for _ in range(times):
            if len(self.points) < 2:
                break
            new_pts = []
            count = len(self.points)
            limit = count if self.closed else count - 1
            for i in range(limit):
                curr_p = self.points[i]
                next_p = self.points[(i + 1) % count]
                mid = (curr_p + next_p) / 2
                new_pts.append(curr_p)
                new_pts.append(mid)
            if not self.closed:
                new_pts.append(self.points[-1])
            self.points = new_pts
        return self.refresh()

    def simplify(self, tolerance: float = 1e-2) -> Self:
        """
        Removes collinear points.
        If the distance of a point from the line segment connecting its neighbors
        is less than 'tolerance', the point is removed.
        """
        if len(self.points) < 3:
            return self

        # Iterative simplification
        # We create a new list and only add points that "matter"
        new_points = [self.points[0]]

        for i in range(1, len(self.points) - 1):
            prev = new_points[-1]  # Look at the last *kept* point
            curr = self.points[i]
            next_p = self.points[i + 1]

            d = Point.distance_to_segment(curr, prev, next_p)

            # If significant deviation, keep it.
            if d > tolerance:
                new_points.append(curr)

        new_points.append(self.points[-1])

        # If closed, check if start/end are collinear with the loop
        # (Omitted for brevity, but essentially check points[0] vs points[-1] and points[1])

        self.points = new_points
        return self.refresh()

    def apply(self, func: Callable[[Point], Point]) -> Self:
        self.points = [func(p) for p in self.points]
        return self

    def _build(self) -> Shape:
        shape = Path(**self._kwargs)
        s = max(0.0, min(1.0, self.smoothness))

        # TODO: Refactor both parts
        if self.closed:
            p_last = self.points[-1]
            p_first = self.points[0]
            start_pt = p_last.lerp(p_first, 0.5)
            shape.jump_to(start_pt.x, start_pt.y)

            n = len(self.points)
            for i in range(n):
                prev = self.points[(i - 1) % n]
                curr = self.points[i]
                next = self.points[(i + 1) % n]
                v_in = curr - prev
                v_out = next - curr
                radius = min(v_in.magnitude(), v_out.magnitude()) / 2.0 * s

                if radius < 1e-6:
                    shape.line_to(curr.x, curr.y)
                else:
                    p_start = curr - v_in.normalize() * radius
                    p_end = curr + v_out.normalize() * radius
                    shape.line_to(p_start.x, p_start.y)
                    shape.quadratic_to(curr.x, curr.y, p_end.x, p_end.y)
            shape.close()

        else:
            shape.jump_to(self.points[0].x, self.points[0].y)
            for i in range(1, len(self.points) - 1):
                prev_p = self.points[i - 1]
                curr_p = self.points[i]
                next_p = self.points[i + 1]
                vec_in = curr_p - prev_p
                vec_out = next_p - curr_p
                radius = min(vec_in.magnitude(), vec_out.magnitude()) / 2.0 * s
                if radius < 1e-6:
                    shape.line_to(curr_p.x, curr_p.y)
                else:
                    p_start = curr_p - vec_in.normalize() * radius
                    p_end = curr_p + vec_out.normalize() * radius
                    shape.line_to(p_start.x, p_start.y)
                    shape.quadratic_to(curr_p.x, curr_p.y, p_end.x, p_end.y)
            shape.line_to(self.points[-1].x, self.points[-1].y)

        return shape

    @classmethod
    def poly(
        cls, n: int, radius: float, orientation: Point | None = None, **kwargs
    ) -> Self:
        """
        Creates a regular polygon centered at (0,0).

        Args:
            n: Number of sides (3=Triangle, 4=Diamond/Square, 6=Hexagon).
            radius: Distance from center to vertex.
            orientation: Vector pointing to the first vertex. Defaults to Point.up.
        """
        if n < 3:
            raise ValueError("Polygon must have at least 3 sides")

        if orientation is None:
            orientation = Point(0, -1)  # Point.up

        # Calculate start angle from the orientation vector
        start_angle = math.atan2(orientation.y, orientation.x)
        step = 2 * math.pi / n

        points = []
        for i in range(n):
            theta = start_angle + i * step
            points.append(Point(radius * math.cos(theta), radius * math.sin(theta)))

        # Create closed polyline
        return cls(points, closed=True, **kwargs)

    def expand(self, delta: float) -> Self:
        """
        Pushes all points away from the origin (0,0) by an absolute amount.
        Useful for making a shape 'thicker' or 'larger' without scaling.

        Note: For best results, ensure the shape is centered first via .center().
        """
        new_points = []
        for p in self.points:
            mag = p.magnitude()
            if mag == 0:
                new_points.append(p)
            else:
                # P_new = P + unit_vector * delta
                # P_new = P + (P / mag) * delta
                # P_new = P * (1 + delta / mag)
                scale_factor = 1 + (delta / mag)
                new_points.append(p * scale_factor)

        self.points = new_points
        return self.refresh()

    def contract(self, delta: float) -> Self:
        """
        Pulls all points towards the origin (0,0) by an absolute amount.
        Negative values will expand.
        """
        return self.expand(-delta)


class Line(Component):
    def __init__(
        self,
        p1: Point | Callable[[], Point],
        p2: Point | Callable[[], Point],
        curvature: float = 0.0,
        fill: Color = Colors.Transparent,
        stroke: Color = Colors.Black,
        width: float = 1.0,
        marker_start: str | None = None,
        marker_end: str | None = None,
    ) -> None:
        super().__init__(
            fill=fill,
            stroke=stroke,
            width=width,
            marker_start=marker_start,
            marker_end=marker_end,
        )
        self.p1 = p1
        self.p2 = p2
        self.curvature = curvature

        if callable(self.p1) or callable(self.p2):
            self._dynamic = True

    def _resolve(self) -> tuple[Point, Point]:
        p1 = self.p1() if callable(self.p1) else self.p1
        p2 = self.p2() if callable(self.p2) else self.p2
        return p1, p2

    def _build(self) -> Shape:
        shape = Path(**self._kwargs)
        start, end = self._resolve()

        shape.jump_to(start.x, start.y)

        if abs(self.curvature) < 1e-4:
            shape.line_to(end.x, end.y)
        else:
            dx = end.x - start.x
            dy = end.y - start.y
            dist = math.sqrt(dx * dx + dy * dy)

            if dist == 0:
                return shape

            s = self.curvature * (dist / 2.0)
            r = (s**2 + (dist / 2.0) ** 2) / (2 * abs(s))
            large_arc = 0
            sweep = 1 if self.curvature > 0 else 0
            shape.arc(r, r, 0, large_arc, sweep, end.x, end.y)

        return shape


class Arrow(Line):
    def __init__(
        self,
        p1: Point | Callable[[], Point],
        p2: Point | Callable[[], Point],
        curvature: float = 0.0,
        fill: Color = Colors.Transparent,
        stroke: Color = Colors.Black,
        width: float = 1.0,
        marker_start: str | None = None,
        marker_end: str | None = "arrow",
    ) -> None:
        super().__init__(
            p1,
            p2,
            curvature,
            fill,
            stroke,
            width,
            marker_start=marker_start,
            marker_end=marker_end,
        )



import math
import pytest
from tesserax.core import Point, Bounds
from tesserax.base import Rect, Circle, Group, Polyline, Text

# --- Fixtures ---


@pytest.fixture
def basic_rect():
    return Rect(100, 50)


@pytest.fixture
def basic_group():
    return Group()


@pytest.fixture
def simple_polyline():
    return Polyline([Point(0, 0), Point(10, 10), Point(20, 0)])


# --- Shape Invariants ---


class TestPrimitives:
    def test_render_contract(self, basic_rect):
        """Invariant: Shapes must produce their corresponding SVG tags."""
        svg = basic_rect._render()
        assert "<rect" in svg
        assert 'width="100"' in svg

        c = Circle(10)
        assert "<circle" in c._render()

    def test_local_dimensionality(self):
        """Invariant: Local bounds match constructor params."""
        w, h = 123.4, 56.7
        r = Rect(w, h)
        b = r.local()
        assert math.isclose(b.width, w)
        assert math.isclose(b.height, h)

        radius = 15
        c = Circle(radius)
        assert math.isclose(c.local().width, radius * 2)

    def test_transform_isolation(self, basic_rect):
        """Invariant: Transform affects global bounds, but NEVER local bounds."""
        initial_local = basic_rect.local()

        # Apply heavy transformations
        basic_rect.translated(100, 100).rotated(math.radians(45)).scaled(2.0)

        # Local bounds must remain immutable
        assert basic_rect.local() == initial_local

        # Global bounds should change
        assert basic_rect.bounds() != initial_local


# --- Group Invariants ---


class TestGroups:
    def test_parental_authority(self, basic_group, basic_rect):
        """Invariant: Adding to group sets parent."""
        assert basic_rect.parent is None
        basic_group.add(basic_rect)
        assert basic_rect.parent is basic_group

    def test_bounds_containment(self):
        """Invariant: Group bounds is union of children."""
        g = Group()
        r1 = Rect(10, 10).translated(0, 0)  # Center at 0,0
        r2 = Rect(10, 10).translated(20, 0)  # Center at 20,0
        g.add(r1, r2)

        # r1 bounds: x[-5, 5]
        # r2 bounds: x[15, 25]
        # Union x: [-5, 25], width=30
        gb = g.local()
        assert math.isclose(gb.x, -5)
        assert math.isclose(gb.width, 30)

    def test_strict_mode_reparenting(self):
        """Invariant: Strict mode prevents stealing children."""
        g1 = Group()
        g2 = Group()
        r = Rect(10, 10)

        g1.add(r)

        with pytest.raises(RuntimeError):
            g2.add(r, mode="strict")


# --- Polyline Invariants ---


class TestPolyline:
    def test_subdivide_monotonicity(self, simple_polyline):
        """Invariant: Subdivision increases resolution."""
        initial_count = len(simple_polyline.points)
        simple_polyline.subdivide(1)
        assert len(simple_polyline.points) > initial_count

    def test_centering_zerosum(self):
        """Invariant: center() moves geometric center to (0,0)."""
        # Create a line far away from origin
        p = Polyline([Point(100, 100), Point(110, 110)])

        p.center()

        # Calculate new centroid
        xs = [pt.x for pt in p.points]
        ys = [pt.y for pt in p.points]
        cx = sum(xs) / len(xs)
        cy = sum(ys) / len(ys)

        assert math.isclose(cx, 0, abs_tol=1e-9)
        assert math.isclose(cy, 0, abs_tol=1e-9)

    def test_topological_closure(self):
        """Invariant: Closed polyline has 'Z' command."""
        p = Polyline([Point(0, 0), Point(1, 1)], closed=True)
        assert "Z" in p._render()



from abc import ABC, abstractmethod
from .core import Body, Point


class Field(ABC):
    @abstractmethod
    def apply(self, bodies: list[Body]):
        pass


class Gravity(Field):
    def __init__(self, g: float = 980.0):  # Pixels/s^2
        self.g = Point(0, g)

    def apply(self, bodies: list[Body]):
        for b in bodies:
            if not b.static:
                # F = m * a -> a is just g
                b.acc += self.g


class Drag(Field):
    def __init__(self, k: float = 0.1):
        self.k = k

    def apply(self, bodies: list[Body]):
        for b in bodies:
            if not b.static:
                # Simple linear drag
                b.acc -= b.vel * self.k
